# Copyright 2021 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compute features from speech signals"""


import abc

from itertools import count
from typing import Mapping, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import numpy as np

from pydrobert.speech.alias import AliasedFactory, alias_factory_subclass_from_arg
from pydrobert.speech import config
from pydrobert.speech.filters import GammaWindow
from pydrobert.speech.filters import HannWindow
from pydrobert.speech.filters import LinearFilterBank
from pydrobert.speech.filters import WindowFunction

__all__ = [
    "frame_by_frame_calculation",
    "FrameComputer",
    "LinearFilterBankFrameComputer",
    "ShortIntegrationFrameComputer",
    "ShortTimeFourierTransformFrameComputer",
    "SIFrameComputer",
    "STFTFrameComputer",
]


class FrameComputer(AliasedFactory):
    """Construct features from a signal from fixed-length segments

    A signal is treated as a (possibly overlapping) time series of frames. Each frame is
    transformed into a fixed-length vector of coefficients.

    Features can be computed one at a time, for example:

    >>> chunk_size = 2 ** 10
    >>> while len(signal):Z
    >>>     segment = signal[:chunk_size]
    >>>     feats = computer.compute_chunk(segment)
    >>>     # do something with feats
    >>>     signal = signal[chunk_size:]
    >>> feats = computer.finalize()

    Or all at once (which can be much faster, depending on how the computer is
    optimized):

    >>> feats = computer.compute_full(signal)

    The k-th frame can be roughly localized to the signal offset to about ``signal[k *
    computer.frame_shift]``. The signal's exact region of influence is dictated by the
    `frame_style` property.
    """

    @abc.abstractproperty
    def frame_style(self) -> Literal["causal", "centered"]:
        """str : Dictates how the signal is split into frames

        If :obj:`'causal'`, the k-th frame is computed over the indices ``signal[k *
        frame_shift:k * frame_shift + frame_length]`` (at most). If :obj:`'centered'`,
        the k-th frame is computed over the indices ``signal[k * frame_shift -
        (frame_length + 1) // 2 + 1:k * frame_shift + frame_length // 2 + 1]``. Any
        range beyond the bounds of the signal is generated in an implementation-specific
        way.
        """
        pass

    @abc.abstractproperty
    def sampling_rate(self) -> float:
        """float : Number of samples in a second of a target recording"""
        pass

    @abc.abstractproperty
    def frame_length(self) -> int:
        """int : Number of samples which dictate a feature vector"""
        pass

    @property
    def frame_length_ms(self) -> float:
        """float : Number of milliseconds of audio which dictate a feature vector"""
        return self.frame_length * 1000 / self.sampling_rate

    @abc.abstractproperty
    def frame_shift(self) -> int:
        """int : Number of samples absorbed between successive frame computations"""
        pass

    @property
    def frame_shift_ms(self) -> float:
        """float : Number of milliseconds between succecssive frame computations"""
        return self.frame_shift * 1000 / self.sampling_rate

    @abc.abstractproperty
    def num_coeffs(self) -> int:
        """int : Number of coefficients returned per frame"""
        pass

    @abc.abstractproperty
    def started(self) -> bool:
        """bool : Whether computations for a signal have started

        Becomes :obj:`True` after the first call to :func:`compute_chunk`. Becomes
        :obj:`False` after call to :func:`finalize`
        """
        pass

    @abc.abstractmethod
    def compute_chunk(self, chunk: np.ndarray) -> np.ndarray:
        """Compute some coefficients, given a chunk of audio

        Parameters
        ----------
        chunk
            A 1D float array of the signal. Should be contiguous and non-overlapping
            with any previously processed segments in the audio stream

        Returns
        -------
        chunk : np.ndarray
            A 2D float array of shape ``(num_frames, num_coeffs)``. ``num_frames`` is
            nonnegative (possibly 0). Contains some number of feature vectors, ordered
            in time over axis 0.
        """
        pass

    @abc.abstractmethod
    def finalize(self) -> np.ndarray:
        """Conclude processing a stream of audio, processing any stored buffer

        Returns
        -------
        chunk : np.ndarray
            A 2D float array of shape ``(num_frames, num_coeffs)``. ``num_frames`` is
            either 1 or 0.
        """
        pass

    def compute_full(self, signal: np.ndarray) -> np.ndarray:
        """Compute a full signal's worth of feature coefficients

        Parameters
        ----------
        signal
            A 1D float array of the entire signal

        Returns
        -------
        spec : np.ndarray
            A 2D float array of shape ``(num_frames, num_coeffs)``. ``num_frames`` is
            nonnegative (possibly 0). Contains some number of feature vectors, ordered
            in time over axis 0.

        Raises
        ------
        ValueError
            If already begin computing frames (``started=True``), and :func:`finalize`
            has not been called
        """
        return frame_by_frame_calculation(self, signal)


class LinearFilterBankFrameComputer(FrameComputer):
    """Frame computers whose features are derived from linear filter banks

    Computers based on linear filter banks have a predictable number of coefficients and
    organization. Like the banks, the features with lower indices correspond to filters
    with lower bandwidths. `num_coeffs` will be simply ``bank.num_filts +
    int(include_energy)``.

    Parameters
    ----------
    bank
        Each filter in the bank corresponds to a coefficient in a frame vector. Can be a
        :class:`LinearFilterBank` or something compatible with
        :func:`pydrobert.speech.alias.alias_factory_subclass_from_arg`
    include_energy
        Whether to include a coefficient based on the energy of the signal within the
        frame. If :obj:`True`, the energy coefficient will be inserted at index 0.
    """

    def __init__(
        self, bank: Union[LinearFilterBank, Mapping, str], include_energy: bool = False
    ):
        self._bank = alias_factory_subclass_from_arg(LinearFilterBank, bank)
        self._include_energy = bool(include_energy)

    @property
    def bank(self) -> LinearFilterBank:
        """LinearFilterBank : The LinearFilterBank from which features are derived"""
        return self._bank

    @property
    def includes_energy(self) -> bool:
        """bool : Whether the first coefficient is an energy coefficient"""
        return self._include_energy

    @property
    def num_coeffs(self) -> int:
        return self._bank.num_filts + int(self._include_energy)


def _power(x):
    return np.linalg.norm(x, ord=2) ** 2


def _mag(x):
    return np.sum(np.abs(x))


class ShortTimeFourierTransformFrameComputer(LinearFilterBankFrameComputer):
    """Compute features of a signal by integrating STFTs

    Computations are per frame and as follows:

    1. The current frame is multiplied with some window (rectangular, Hamming, Hanning,
       etc)
    2. A DFT is performed on the result
    3. For each filter in the provided input bank:

       a. Multiply the result of 2. with the frequency response of the filter
       b. Sum either the pointwise square or absolute value of elements in the buffer
          from 3a.
       c. Optionally take the log of the sum

    Warning
    --------
    This behaviour differs from that of [povey2011]_ or [young]_ in three ways. First,
    the sum (3b) comes after the filtering (3a), which changes the result in the squared
    case. Second, the sum is over the full power spectrum, rather than just between 0
    and the Nyquist. This doubles the value at the end of 3c. if a real filter is used.
    Third, frame boundaries are calculated diffferently.

    Parameters
    ----------
    bank
        Each filter in the bank corresponds to a coefficient in a frame vector. Can be a
        :class:`LinearFilterBank` or something compatible with
        :func:`pydrobert.speech.alias.alias_factory_subclass_from_arg`
    frame_length_ms
        The length of a frame, in milliseconds. Defaults to the length of the largest
        filter in the bank
    frame_shift_ms : float, optional
        The offset between successive frames, in milliseconds
    frame_style
        Defaults to :obj:`'centered'` if ``bank.is_zero_phase``, :obj:`'causal'`
        otherwise.
    include_energy
    pad_to_nearest_power_of_two
        Whether the DFT should be a padded to a power of two for computational
        efficiency
    window_function
        The window used in step 1. Can be a :class:`WindowFunction` or something
        compatible with :func:`pydrobert.speech.alias_factory_subclass_from_arg`.
        Defaults to :class:`pydrobert.speech.filters.GammaWindow` when `frame_style` is
        :obj:`'causal'`, otherwise :class:`pydrobert.speech.filters.HannWindow`.
    use_log
        Whether to take the log of the sum from 3b.
    use_power
        Whether to sum the power spectrum or the magnitude spectrum
    kaldi_shift
        If :obj:`True`, the k-th frame will be computed using the signal between
        ``signal[ k - frame_length // 2 + frame_shift // 2:k + (frame_length + 1) // 2
        + frame_shift // 2]``. These are the frame bounds for Kaldi [povey2011]_.
    """

    aliases = {"stft"}  #:

    def __init__(
        self,
        bank: Union[LinearFilterBank, Mapping, str],
        frame_length_ms: Optional[float] = None,
        frame_shift_ms: Optional[float] = 10,
        frame_style: Optional[Literal["causal", "centered"]] = None,
        include_energy: bool = False,
        pad_to_nearest_power_of_two: bool = True,
        window_function: Optional[Union[WindowFunction, Mapping, str]] = None,
        use_log: bool = True,
        use_power: bool = False,
        kaldi_shift: bool = False,
    ):
        bank = alias_factory_subclass_from_arg(LinearFilterBank, bank)
        self._rate = bank.sampling_rate
        self._frame_shift = int(0.001 * frame_shift_ms * self._rate)
        self._log = use_log
        self._power = use_power
        self._real = bank.is_real
        self._started = False
        self._first_frame = True
        self._buf_len = 0
        self._chunk_dtype = np.float64
        self._kaldi_shift = kaldi_shift
        if frame_style is None:
            frame_style = "centered" if bank.is_zero_phase else "causal"
        elif frame_style not in ("centered", "causal"):
            raise ValueError('Invalid frame style: "{}"'.format(frame_style))
        self._frame_style = frame_style
        if frame_length_ms is None:
            self._frame_length = max(
                max(right - left for left, right in bank.supports),
                # ensure at least one dft bin is nonzero per filter
                int(
                    np.ceil(
                        2
                        * self._rate
                        / min(right - left for left, right in bank.supports_hz)
                    )
                ),
            )
        else:
            self._frame_length = int(0.001 * frame_length_ms * bank.sampling_rate)
        self._buf = np.empty(self._frame_length, dtype=np.float64)
        if window_function is None:
            if frame_style == "causal":
                window_function = GammaWindow()
            else:
                window_function = HannWindow()
        else:
            window_function = alias_factory_subclass_from_arg(
                WindowFunction, window_function
            )
        self._window = window_function.get_impulse_response(self._frame_length)
        if pad_to_nearest_power_of_two:
            self._dft_size = int(2 ** np.ceil(np.log2(self._frame_length)))
        else:
            self._dft_size = self._frame_length
        if self._power:
            self._nonlin_op = _power
        else:
            self._nonlin_op = _mag
        self._truncated_filts = []
        self._filt_start_idxs = []
        for filt_idx in range(bank.num_filts):
            start_idx, truncated_filt = bank.get_truncated_response(
                filt_idx, self._dft_size
            )
            self._filt_start_idxs.append(start_idx)
            self._truncated_filts.append(truncated_filt)
        super(ShortTimeFourierTransformFrameComputer, self).__init__(
            bank, include_energy=include_energy
        )

    @property
    def frame_style(self) -> str:
        return self._frame_style

    @property
    def sampling_rate(self) -> float:
        return self._rate

    @property
    def frame_length(self) -> int:
        return self._frame_length

    @property
    def frame_shift(self) -> int:
        return self._frame_shift

    @property
    def started(self) -> bool:
        return self._started

    @property
    def kaldi_shift(self) -> bool:
        return self._kaldi_shift

    def _compute_frame(self, frame, coeffs):
        # given the frame, store feature values within coeff
        assert len(frame) == self._frame_length
        assert len(coeffs) == self.num_coeffs
        if self.includes_energy:
            coeffs[0] = np.inner(frame, frame) / self._frame_length
            if not self._power:
                coeffs[0] **= 0.5
            if self._log:
                coeffs[0] = np.log(max(coeffs[0], config.LOG_FLOOR_VALUE))
            coeffs = coeffs[1:]
        if config.USE_FFTPACK:
            from scipy import fftpack

            is_odd = self._dft_size % 2
            buffered_frame = np.zeros(self._dft_size + 2 - is_odd, dtype=np.float64)
            buffered_frame[1 : self._frame_length + 1] = frame
            buffered_frame[1 : self._frame_length + 1] *= self._window
            buffered_frame[1 : self._dft_size + 1] = fftpack.rfft(
                buffered_frame[1 : self._dft_size + 1], overwrite_x=True
            )
            buffered_frame[0] = buffered_frame[1]
            buffered_frame[1] = 0
            half_spect = buffered_frame.view(np.complex128)
        else:
            half_spect = np.fft.rfft(frame * self._window, n=self._dft_size)
        assert half_spect.dtype == np.complex128
        half_len = len(half_spect)
        for filt_idx in range(len(self._filt_start_idxs)):
            start_idx = self._filt_start_idxs[filt_idx]
            truncated_filt = self._truncated_filts[filt_idx]
            trunc_len = len(truncated_filt)
            consumed = 0
            conjugate = False
            val = 0
            while consumed < trunc_len:
                if conjugate:
                    seg_len = (
                        min(
                            start_idx + trunc_len - consumed,
                            half_len - 2 + half_len % 2,
                        )
                        - start_idx
                    )
                    seg_len = max(0, seg_len)
                    if seg_len:
                        val += self._nonlin_op(
                            half_spect[
                                (-2 + (half_len % 2) - start_idx) : (
                                    -2 + (half_len % 2) - start_idx - seg_len
                                ) : -1
                            ].conj()
                            * truncated_filt[consumed : consumed + seg_len]
                        )
                    start_idx -= half_len - 2 + half_len % 2
                else:
                    seg_len = min(start_idx + trunc_len - consumed, half_len)
                    seg_len -= start_idx
                    seg_len = max(0, seg_len)
                    if seg_len:
                        val += self._nonlin_op(
                            half_spect[start_idx : start_idx + seg_len]
                            * truncated_filt[consumed : consumed + seg_len]
                        )
                    start_idx -= half_len
                conjugate = not conjugate
                consumed += seg_len
                start_idx = max(0, start_idx)
            if self._real:
                val *= 2
            if self._log:
                val = np.log(max(val, config.LOG_FLOOR_VALUE))
            coeffs[filt_idx] = val

    def compute_chunk(self, chunk: np.ndarray) -> np.ndarray:
        self._chunk_dtype = chunk.dtype  # needed for `finalize`
        # algorithm should work when frame shift is greater than frame
        # length - buf_len may be negative, which will skip samples
        buf_len = self._buf_len
        chunk_len = len(chunk)
        total_len = chunk_len + buf_len
        noncausal_first = self._frame_style == "centered"
        noncausal_first &= self._first_frame
        if noncausal_first:
            if self._kaldi_shift:
                frame_length = (self._frame_length + 1) // 2
                frame_length += self._frame_shift // 2
            else:
                frame_length = self._frame_length // 2 + 1
        else:
            frame_length = self._frame_length
        frame_shift = self._frame_shift
        num_frames = max(0, (total_len - frame_length) // frame_shift + 1)
        coeffs = np.empty((num_frames, self.num_coeffs), dtype=self._chunk_dtype)
        for frame_idx in range(num_frames):
            frame_start_idx = frame_idx * frame_shift
            if frame_start_idx < buf_len:
                frame = np.concatenate(
                    [
                        self._buf[-(buf_len - frame_start_idx) :],
                        chunk[: frame_length - buf_len + frame_start_idx],
                    ]
                )
            else:
                frame = chunk[
                    (frame_start_idx - buf_len) : (
                        frame_start_idx - buf_len + frame_length
                    )
                ]
            if noncausal_first:
                # the first frame's l.h.s is a reflection of its r.h.s.
                # shove the reflection into the buf - later frames
                # might need it
                chunk = chunk[(frame_length - buf_len) :]
                chunk_len -= frame_length - buf_len
                frame_length = self._frame_length
                if self._kaldi_shift:
                    self._buf[:] = np.pad(
                        frame,
                        (self._frame_length // 2 - self._frame_shift // 2, 0),
                        "symmetric",
                    )
                else:
                    self._buf[:] = np.pad(
                        frame, ((frame_length + 1) // 2 - 1, 0), "symmetric"
                    )
                frame = self._buf
                total_len = chunk_len + frame_length
                buf_len = frame_length
                noncausal_first = False
            self._compute_frame(frame, coeffs[frame_idx])
            self._first_frame = False
        rem_len = total_len - num_frames * frame_shift
        assert rem_len < frame_length
        if rem_len > 0:
            throw_away = total_len - rem_len
            if throw_away < buf_len:
                rem_ring_len = buf_len - throw_away
                assert rem_ring_len < rem_len or (
                    rem_ring_len <= rem_len and not len(chunk)
                )
                self._buf[
                    self._frame_length
                    - rem_len : self._frame_length
                    - rem_len
                    + rem_ring_len
                ] = self._buf[self._frame_length - rem_ring_len :]
                self._buf[self._frame_length - (rem_len - rem_ring_len) :] = chunk
            else:
                self._buf[-rem_len:] = chunk[-rem_len:]
        self._buf_len = rem_len
        self._started = True
        return coeffs

    def finalize(self) -> np.ndarray:
        buf_len = self._buf_len
        frame_length = self._frame_length
        frame_shift = self._frame_shift
        if self._frame_style == "causal":
            pad_left = 0
        elif self._kaldi_shift:
            pad_left = frame_length // 2 - frame_shift // 2
        else:
            pad_left = (frame_length + 1) // 2 - 1
        num_frames = buf_len + frame_shift // 2
        if not self._first_frame:
            num_frames -= pad_left
            pad_left = 0
        num_frames //= frame_shift
        if num_frames >= 1:
            pad_right = (num_frames - 1) * frame_shift + frame_length - buf_len
            pad_right -= pad_left
            coeffs = np.empty((num_frames, self.num_coeffs), dtype=self._chunk_dtype)
            frames = np.pad(self._buf[-buf_len:], (pad_left, pad_right), "symmetric",)
            for frame_idx in range(num_frames):
                frame = frames[
                    frame_idx * frame_shift : frame_idx * frame_shift + frame_length
                ]
                self._compute_frame(frame, coeffs[frame_idx])
        else:
            coeffs = np.empty((0, self.num_coeffs), dtype=self._chunk_dtype)
        self._buf_len = 0
        self._started = False
        self._first_frame = True
        return coeffs

    def compute_full(self, signal: np.ndarray) -> np.ndarray:
        if self.started:
            raise ValueError("Already started computing frames")
        # there should be a nicer way to calculate this
        frame_length = self._frame_length
        frame_shift = self._frame_shift
        if len(signal) < frame_length // 2 + 1:
            return np.empty((0, self.num_coeffs), dtype=signal.dtype)
        if self._frame_style == "causal":
            pad_left = 0
        elif self._kaldi_shift:
            pad_left = frame_length // 2 - frame_shift // 2
        else:
            pad_left = (self._frame_length + 1) // 2 - 1
        # total_len = pad_left + len(signal)
        # num_frames = max(0, (total_len - frame_length) // frame_shift + 1)
        # rem_len = total_len - num_frames * frame_shift
        # if rem_len >= frame_length // 2 + 1:
        #     num_frames += 1
        #     pad_right = frame_length - rem_len
        # else:
        #     pad_right = 0
        num_frames = max(0, (len(signal) + frame_shift // 2) // frame_shift)
        total_len = (num_frames - 1) * frame_shift - pad_left + frame_length
        pad_right = max(0, total_len - len(signal))
        if pad_left or pad_right:
            signal = np.pad(signal, (pad_left, pad_right), "symmetric")
        coeffs = np.zeros((num_frames, self.num_coeffs), dtype=signal.dtype)
        for frame_idx in range(num_frames):
            frame_left = frame_idx * frame_shift
            self._compute_frame(
                signal[frame_left : frame_left + frame_length], coeffs[frame_idx]
            )
        return coeffs


STFTFrameComputer = ShortTimeFourierTransformFrameComputer


class ShortIntegrationFrameComputer(LinearFilterBankFrameComputer):
    """Compute features by integrating over the filter modulus

    Each filter in the bank is convolved with the signal. A pointwise nonlinearity
    pushes the frequency band towards zero. Most of the energy of the signal can be
    captured in a short time integration. Though best suited to processing whole
    utterances at once, short integration is compatable with the frame analogy if the
    frame is assumed to be the cone of influence of the maximum-length filter.

    For computational purposes, each filter's impulse response is clamped to zero
    outside the support of the largest filter in the bank, making it a finite impulse
    response filter. This effectively decreases the frequency resolution of the filters
    which aren't already FIR. For better frequency resolution at the cost of
    computational time, increase
    :obj:`pydrobert.speech.config.EFFECTIVE_SUPPORT_THRESHOLD`.

    Parameters
    ----------
    bank
        Each filter in the bank corresponds to a coefficient in a frame vector. Can be a
        :class:`LinearFilterBank` or something compatible with
        :func:`pydrobert.speech.alias.alias_factory_subclass_from_arg`
    frame_shift_ms
        The offset between successive frames, in milliseconds. Also the length of the
        integration
    frame_style
        Defaults to :obj:`'centered'` if `bank.is_zero_phase`, :obj:`'causal'`
        otherwise. If :obj:`'centered'` each filter of the bank is translated so that
        its support lies in the center of the frame
    include_energy
    pad_to_nearest_power_of_two
        Pad the DFTs used in computation to a power of two for efficient computation
    window_function : pydrobert.speech.filters.WindowFunction, dict, or str
        The window used to weigh integration. Can be a :class:`WindowFunction` or
        something compatible with
        :func:`pydrobert.speech.alias_factory_subclass_from_arg`. Defaults to
        :class:`pydrobert.speech.filters.GammaWindow` when ``frame_style`` is
        :obj:`'causal'`, otherwise :class:`pydrobert.speech.filters.HannWindow`.
    use_power
        Whether the pointwise linearity is the signal's power or magnitude
    use_log
        Whether to take the log of the integration
    """

    aliases = {"si"}  #:

    def __init__(
        self,
        bank: Union[LinearFilterBank, Mapping, str],
        frame_shift_ms: float = 10,
        frame_style: Optional[Literal["causal", "centered"]] = None,
        include_energy: bool = False,
        pad_to_nearest_power_of_two: bool = True,
        window_function: Optional[Union[WindowFunction, Mapping, str]] = None,
        use_power: bool = False,
        use_log: bool = True,
    ):
        bank = alias_factory_subclass_from_arg(LinearFilterBank, bank)
        self._rate = bank.sampling_rate
        self._frame_shift = int(0.001 * frame_shift_ms * self._rate)
        self._log = bool(use_log)
        self._power = bool(use_power)
        self._real = bank.is_real
        self._ret_dtype = np.float64
        self._x_rem, self._y_rem, self._skip = 0, 0, 0
        self._started = False
        if frame_style is None:
            frame_style = "centered" if bank.is_zero_phase else "causal"
        elif frame_style not in ("centered", "causal"):
            raise ValueError('Invalid frame style: "{}"'.format(frame_style))
        self._frame_style = frame_style
        if window_function is None:
            if frame_style == "causal":
                window_function = GammaWindow()
            else:
                window_function = HannWindow()
        else:
            window_function = alias_factory_subclass_from_arg(
                WindowFunction, window_function
            )
        window = window_function.get_impulse_response(2 * self._frame_shift)
        self._window = window.reshape(2, self._frame_shift)
        if frame_style == "centered":
            # we will recenter all filters so that their zero sample
            # is at max_support // 2
            self._max_support = max(right - left for left, right in bank.supports)
            self._translation = self._max_support // 2
        else:
            # we will shift all filters by whatever the minimum value
            # makes them all supported above/equal 0. We treat all that
            # translated space as nonzero for the sake of the
            # overlap-add algorithm
            self._translation = 0
            self._max_support = 0
            for left, right in bank.supports:
                self._translation = max(-left, self._translation)
                self._max_support = max(self._max_support, right)
            self._max_support += self._translation
        min_support_hz = min(right - left for left, right in bank.supports_hz)
        self._frame_length = self._max_support + self._frame_shift - 1
        self._dft_size = max(
            self._frame_length,
            # make sure the effective support is represented in at least
            # one dft bin
            int(np.ceil(2 * self._rate / min_support_hz)),
        )
        if pad_to_nearest_power_of_two:
            self._dft_size = int(2 ** np.ceil(np.log2(self._dft_size)))
        self._x_buf = np.empty(self._dft_size, dtype=np.float64)
        self._filts = []
        if include_energy:
            # this is a hacky way to make sure we get an accurate energy
            # coefficient - dirac deltas return the signal or a
            # translation of it
            dirac_filter = np.zeros(self._dft_size, dtype=np.float64)
            dirac_filter[self._translation] = 1
            if self._real:
                dirac_filter = np.fft.rfft(dirac_filter)
            else:
                dirac_filter = np.fft.fft(dirac_filter)
            self._filts.append(dirac_filter)
        for filt_idx in range(bank.num_filts):
            filt = bank.get_impulse_response(filt_idx, self._dft_size)
            if frame_style == "centered":
                left_samp, right_samp = bank.supports[filt_idx]
                mid_samp = (left_samp + right_samp) // 2
                filt = np.roll(filt, self._translation - mid_samp + 1)
            else:
                filt = np.roll(filt, self._translation)
            # we clamp the support in time to make the filter FIR.
            self._filts.append(self._compute_dft(filt[: self._max_support]))
        # we don't have to store the filtered signal, just the values
        # that are accumulated in each frame shift. Since integration windows
        # are not in general uniform, we add an index for taking the dot
        # product of the first and second half of the window
        y_blocks = self._dft_size - self._max_support + 2 * self._frame_shift
        y_blocks = int(np.ceil(y_blocks / self._frame_shift))
        self._y_buf = np.empty((y_blocks, 2, len(self._filts)), dtype=np.float64)
        super(ShortIntegrationFrameComputer, self).__init__(
            bank, include_energy=include_energy
        )

    @property
    def frame_style(self) -> str:
        return self._frame_style

    @property
    def sampling_rate(self) -> float:
        return self._rate

    @property
    def frame_length(self) -> int:
        return self._frame_length

    @property
    def frame_shift(self):
        return self._frame_shift

    @property
    def started(self) -> bool:
        return self._started

    def compute_chunk(self, chunk: np.ndarray) -> np.ndarray:
        self._compute_preamble(chunk)
        chunk = self._handle_skip(chunk)
        chunk_len = len(chunk)
        valid_samples_per_dft = self._dft_size - self._max_support + 1
        num_raw = self._x_rem + chunk_len
        num_dfts = num_raw // valid_samples_per_dft
        num_frames = max(0, (num_raw + self._y_rem) // self._frame_shift - 1)
        if num_frames:
            num_processed = (num_frames + 1) * self._frame_shift
        else:
            num_processed = self._y_rem
        if num_processed - self._y_rem > num_dfts * valid_samples_per_dft:
            num_dfts += 1
        coeffs = np.empty((num_frames, self.num_coeffs), dtype=self._ret_dtype)
        cur_frame, chunk_copied = 0, 0
        for dft_idx in range(num_dfts):
            end_idx = min(
                (dft_idx + 1) * valid_samples_per_dft - self._x_rem, chunk_len
            )
            assert end_idx >= 0
            y_keep = end_idx - dft_idx * valid_samples_per_dft + self._x_rem
            start_idx = end_idx - self._dft_size  # relative to chunk
            if start_idx < 0:
                chunk_to_copy = end_idx - chunk_copied
                assert chunk_to_copy < self._dft_size
                self._x_buf[: self._dft_size - chunk_to_copy] = self._x_buf[
                    chunk_to_copy:
                ]
                self._x_buf[self._dft_size - chunk_to_copy :] = chunk[
                    chunk_copied:end_idx
                ]
                chunk_copied = end_idx
                cur_buf = self._x_buf
            else:
                cur_buf = chunk[start_idx:end_idx]
            X_buf = self._compute_dft(cur_buf)
            self._fill_y_buf(X_buf, y_keep)
            del X_buf
            while self._y_rem >= 2 * self._frame_shift:
                self._compute_frame(coeffs[cur_frame, :])
                cur_frame += 1
        assert cur_frame == num_frames, (cur_frame, num_frames)
        if chunk_len - chunk_copied:
            chunk_to_copy = min(self._dft_size, chunk_len - chunk_copied)
            self._x_buf[:-chunk_to_copy] = self._x_buf[chunk_to_copy:]
            self._x_buf[-chunk_to_copy:] = chunk[-chunk_to_copy:]
        self._x_rem = max(0, num_raw - num_dfts * valid_samples_per_dft)
        return coeffs

    def finalize(self) -> np.ndarray:
        coeffs = np.empty((0, self.num_coeffs), dtype=self._ret_dtype)
        if self._started:
            frame_shift = self._frame_shift
            frame_length = self._frame_length
            if self._frame_style == "centered":
                # we 'borrowed' a half frame's worth of coefficients
                # from the start of the sequence in order to center the
                # integration, so we discount that from the remaining
                # samples
                borrowed = frame_shift
            else:
                borrowed = 0
            buf_len = self._translation - self._skip + self._x_rem
            buf_len += self._y_rem - borrowed
            num_frames = max(0, (buf_len + frame_shift // 2) // frame_shift)
            if num_frames >= 1:
                pad_right = (num_frames - 1) * frame_shift + frame_length
                pad_right -= buf_len
                coeffs = self.compute_chunk(np.zeros(pad_right, dtype=self._ret_dtype))[
                    :num_frames
                ]
        self._started = False
        return coeffs

    def compute_full(self, signal: np.ndarray) -> np.ndarray:
        if self._started:
            raise ValueError("Already started computing frames")
        return np.concatenate([self.compute_chunk(signal), self.finalize()])

    def _compute_preamble(self, chunk):
        # check for data type consistency, handle stuff if just started
        if self._started:
            if chunk.dtype != self._ret_dtype:
                raise ValueError("Chunk does not share a type with previous chunks")
        else:
            if not np.issubdtype(chunk.dtype, np.floating):
                raise ValueError("Chunk must be a float type")
            self._ret_dtype = chunk.dtype
            self._x_buf.fill(0)
            self._y_buf.fill(0)
            self._x_rem = 0
            self._y_rem = 0
            if self._frame_style == "centered":
                self._skip = self._translation - self._frame_shift
                if self._skip < 0:
                    self._x_rem = -self._skip
                    self._skip = 0
            else:
                self._skip = self._translation
            self._started = True

    def _handle_skip(self, chunk):
        # 'skip' refers to some number of samples at the beginning of
        # the utterance that won't count towards frames. We add them
        # to x_buf, but do not increment x_rem
        if not self._skip:
            return chunk
        assert not self._x_rem
        consumed = min(self._skip, len(chunk))
        x_len = len(self._x_buf)
        if consumed < x_len:
            self._x_buf[: x_len - consumed] = self._x_buf[consumed:]
            self._x_buf[x_len - consumed :] = chunk[:consumed]
        else:
            self._x_buf[:] = chunk[consumed - x_len : consumed]
        self._skip -= consumed
        return chunk[consumed:]

    def _fill_y_buf(self, X_buf, y_keep):
        # using the fourier domain of the raw signal window, compute
        # convolutions and store accumulations in y_buf
        block_offs = self._y_rem // self._frame_shift
        second_block_start = (block_offs + 1) * self._frame_shift - self._y_rem
        for filt_idx in range(self.num_coeffs):
            Y_buf = X_buf * self._filts[filt_idx]
            y_valid = self._compute_idft(Y_buf)[-y_keep:]
            if self._power:
                y_valid[:] = y_valid * y_valid.conj()
            else:
                y_valid[:] = np.abs(y_valid)
            del Y_buf
            for block_end, block_idx in zip(
                list(
                    range(
                        second_block_start,
                        y_keep + self._frame_shift,
                        self._frame_shift,
                    )
                ),
                count(block_offs),
            ):
                active_end = min(block_end, y_keep)
                active_start = max(0, block_end - self._frame_shift)
                y_active = y_valid[active_start:block_end].real
                window_start = max(0, self._frame_shift - block_end)
                window_end = self._frame_shift - block_end + active_end
                window_active = self._window[:, window_start:window_end]
                # block_accum = np.sum(y_active * window_active, axis=1)
                self._y_buf[block_idx, :, filt_idx] += np.sum(
                    y_active * window_active, axis=1
                )
        self._y_rem += y_keep

    def _compute_dft(self, buff):
        # given a buffer, compute its fourier transform. Always copies
        # the data
        assert len(buff) <= self._dft_size
        if config.USE_FFTPACK and self._real:
            from scipy import fftpack

            buffered_frame = np.zeros(
                self._dft_size + 2 - self._dft_size % 2, dtype=np.float64
            )
            buffered_frame[1 : len(buff) + 1] = buff
            buffered_frame[1 : self._dft_size + 1] = fftpack.rfft(
                buffered_frame[1 : self._dft_size + 1], overwrite_x=True
            )
            buffered_frame[0] = buffered_frame[1]
            buffered_frame[1] = 0
            fourier_frame = buffered_frame.view(np.complex128)
        elif self._real:
            fourier_frame = np.fft.rfft(buff, n=self._dft_size)
        elif config.USE_FFTPACK:
            from scipy import fftpack

            complex_frame = np.zeros(self._dft_size, dtype=np.complex128)
            complex_frame[: len(buff)] = buff  # implicit upcast if f32
            fourier_frame = fftpack.fft(complex_frame, overwrite_x=True)
        else:
            fourier_frame = np.fft.fft(buff, n=self._dft_size)
        assert fourier_frame.dtype == np.complex128
        return fourier_frame

    def _compute_idft(self, fourier_buff):
        # given a buffer, compute its inverse fourier transform. Assume
        # it's ok to modify the buffer.
        assert fourier_buff.dtype == np.complex128
        if config.USE_FFTPACK and self._real:
            from scipy import fftpack

            fourier_buff = fourier_buff.view(np.float64)
            fourier_buff[1] = fourier_buff[0]
            if self._dft_size % 2:
                fourier_buff = fourier_buff[1:]
            else:
                fourier_buff = fourier_buff[1:-1]
            idft = fftpack.irfft(fourier_buff, overwrite_x=True)
        elif self._real:
            idft = np.fft.irfft(fourier_buff, n=self._dft_size)
        elif config.USE_FFTPACK:
            from scipy import fftpack

            idft = fftpack.ifft(fourier_buff, overwrite_x=True)
        else:
            idft = np.fft.ifft(fourier_buff)
        return idft

    def _compute_frame(self, coeffs):
        # compute a frame's worth of coefficients from y_rem
        assert self._y_rem >= 2 * self._frame_shift
        # y_buf[0, 0, :] contains the accumulators of the first half of
        # the frame (the first block) multiplied with the first half of
        # the window
        # y_buf[1, 1, :] contains the accumulators of the second half of
        # the frame (the second block) multiplied with the second half
        # of the window
        coeffs[:] = self._y_buf[0, 0, :] + self._y_buf[1, 1, :]
        if self._log:
            coeffs[:] = np.log(np.maximum(coeffs, config.LOG_FLOOR_VALUE))
        self._y_buf[:-1] = self._y_buf[1:]
        self._y_buf[-1] = 0
        self._y_rem -= self._frame_shift


SIFrameComputer = ShortIntegrationFrameComputer


def frame_by_frame_calculation(
    computer: FrameComputer, signal: np.ndarray, chunk_size: int = 2 ** 10
):
    """Compute feature representation of entire signal iteratively

    This function constructs a feature matrix of a signal through successive calls to
    ``computer.compute_chunk``. Its return value should be identical to that of calling
    ``computer.compute_full(signal)``, but is possibly much slower.
    ``computer.compute_full`` should be favoured.

    Parameters
    ----------
    computer
    signal
        A 1D float array of the entire signal
    chunk_size
        The length of the signal buffer to process at a given time

    Returns
    -------
    spec : np.ndarray
        A 2D float array of shape ``(num_frames, num_coeffs)``. ``num_frames`` is
        nonnegative (possibly 0). Contains some number of feature vectors, ordered in
        time over axis 0.

    Raises
    ------
    ValueError
        If already begin computing frames (``computer.started == True``)
    """
    if computer.started:
        raise ValueError("Already started computing frames")
    coeffs = []
    while len(signal):
        coeffs.append(computer.compute_chunk(signal[:chunk_size]))
        signal = signal[chunk_size:]
    coeffs.append(computer.finalize())
    return np.concatenate(coeffs)



# Copyright 2021 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes for post-processing feature matrices"""


import abc
from typing import Callable, Optional, Union
import warnings

from itertools import count

import numpy as np

from pydrobert.speech.alias import AliasedFactory
from pydrobert.speech.util import read_signal

__all__ = [
    "CMVN",
    "Deltas",
    "PostProcessor",
    "Stack",
    "Standardize",
]


class PostProcessor(AliasedFactory):
    """A container for post-processing features with a transform"""

    @abc.abstractmethod
    def apply(
        self, features: np.ndarray, axis: int = -1, in_place: bool = False
    ) -> np.ndarray:
        """Applies the transformation to a feature tensor

        Consult the class documentation for more details on what the transformation is.

        Parameters
        ----------
        features
        axis
            The axis of `features` to apply the transformation along
        in_place
            Whether it is okay to modify features (:obj:`True`) or whether a copy should
            be made (:obj:`False`)

        Returns
        -------
        out : np.ndarray
            The transformed features
        """
        pass


class Standardize(PostProcessor):
    """Standardize each feature coefficient

    Though the exact behaviour of an instance varies according to below, the "goal" of
    this transformation is such that every feature coefficient on the chosen axis has
    mean 0 and variance 1 (if `norm_var` is :obj:`True`) over the other axes. Features
    are assumed to be real; the return data type after :func:`apply` is always a 64-bit
    float.

    If `rfilename` is not specified or the associated file is empty, coefficients are
    standardized locally (within the target tensor). If `rfilename` is specified,
    feature coefficients are standardized according to the sufficient statistics
    collected in the file. The latter implementation is based off [povey2011]_. The
    statistics will be loaded with :func:`pydrobert.speech.util.read_signal`.

    Parameters
    ----------
    rfilename
    norm_var
    **kwargs

    Notes
    -----
    Additional keyword arguments can be passed to the initializer if rfilename is set.
    They will be passed on to :func:`pydrobert.speech.util.read_signal`

    See Also
    --------
    pydrobert.speech.util.read_signal
        Describes the strategy used for loading signals
    """

    aliases = {"standardize", "normalize", "unit", "cmvn"}  #:

    def __init__(
        self, rfilename: Optional[str] = None, norm_var: bool = True, **kwargs
    ):
        self._stats = None
        self._norm_var = bool(norm_var)
        if rfilename is not None:
            if "dtype" in kwargs:
                self._stats = read_signal(rfilename, **kwargs)
            else:
                for dtype in (np.float64, np.float32, "dm", "fm"):
                    try:
                        self._stats = read_signal(rfilename, dtype=dtype, **kwargs)
                        break
                    except (IOError, ValueError, ImportError, TypeError):
                        pass
                if self._stats is None:
                    raise IOError("Unable to load stats from {}".format(rfilename))
                if len(self._stats.shape) == 1:
                    # stats were likely stored as simple binary. Need
                    # to make sure we've cast to the right kind of
                    # float. Probably a non-issue if we saved the data
                    # ourselves
                    self._sanitize_stats()
        elif kwargs:
            raise TypeError("Invalid keyword arguments: {}".format(tuple(kwargs)))
        super(Standardize, self).__init__()

    def _sanitize_stats(self, checked_other_float: bool = False):
        try:
            self._stats = self._stats.reshape((2, -1))
            valid = np.isclose(np.round(self._stats[0, -1]), self._stats[0, -1])
            valid &= np.all(self._stats >= 0)
        except ValueError:
            # in this case we couldn't reshape to (2, -1).
            valid = False
        if not valid and checked_other_float:
            raise IOError(
                "Could not properly load statistics. Try specifying "
                "additional parameters in init (see docstring)"
            )
        elif not valid:
            if self._stats.dtype not in (np.float32, np.float64):
                raise ValueError(
                    "Statistics were loaded with a weird data type ({}) and "
                    "are invalid. Make sure the arguments you passed to "
                    "the init are correct".format(self._stats.dtype)
                )
            elif self._stats.dtype == np.float32:
                self._stats = np.frombuffer(self._stats.tobytes(), dtype=np.float64)
            else:
                self._stats = np.frombuffer(
                    self._stats.tobytes(), dtype=np.float32
                ).astype(np.float64)
            self._sanitize_stats(True)

    @property
    def have_stats(self) -> bool:
        """bool : Whether at least one feature vector has been accumulated"""
        return self._stats is not None and self._stats[0, -1]

    def _accumulate_vector(self, vec):
        # accumulate over a single feature vector
        num_coeffs = len(vec)
        if self._stats is None:
            self._stats = np.zeros((2, num_coeffs + 1), dtype=np.float64)
        elif self._stats.shape[1] != num_coeffs + 1:
            raise ValueError(
                "Expected feature vector of length {}; got {}".format(
                    self._stats.shape[1] - 1, num_coeffs
                )
            )
        self._stats[0, -1] += 1
        self._stats[0, :-1] += vec.astype(np.float64)
        self._stats[1, :-1] += np.square(vec, dtype=np.float64)

    def _accumulate_tensor(self, tensor, axis):
        # accumulate over a tensor (with a shape)
        num_coeffs = tensor.shape[axis]
        if self._stats is None:
            self._stats = np.zeros((2, num_coeffs + 1), dtype=np.float64)
        elif self._stats.shape[1] != num_coeffs + 1:
            raise ValueError(
                "Expected feature vector of length {}; got {}".format(
                    self._stats.shape[1] - 1, num_coeffs
                )
            )
        other_axes = tuple(
            idx for idx in range(len(tensor.shape)) if idx != axis % len(tensor.shape)
        )
        self._stats[0, -1] += np.prod(tuple(tensor.shape[idx] for idx in other_axes))
        self._stats[0, :-1] += tensor.sum(axis=other_axes, dtype=np.float64)
        self._stats[1, :-1] += np.square(tensor, dtype=np.float64).sum(axis=other_axes)

    def accumulate(self, features: np.ndarray, axis: int = -1) -> None:
        """Accumulate statistics from a feature tensor

        Parameters
        ----------
        features
        axis

        Raises
        ------
        ValueError
            If the length of `axis` does not match that of past feature
            vector lengths
        """
        if (features.shape and not np.prod(features.shape)) or not len(features):
            raise ValueError("Cannot accumulate from empty array")
        if features.shape and features.ndim > 1:
            self._accumulate_tensor(features, axis)
        else:
            self._accumulate_vector(features)

    def _apply_vector(self, vec, in_place):
        # apply transformation to vector
        num_coeffs = len(vec)
        if self._stats is not None and self._stats.shape[1] != num_coeffs + 1:
            raise ValueError(
                "Expected feature vector of length {}; got {}".format(
                    self._stats.shape[1] - 1, num_coeffs
                )
            )
        if not in_place or vec.dtype != np.float64:
            vec = vec.astype(np.float64)
        if self.have_stats:
            count = self._stats[0, -1]
            means = self._stats[0, :-1] / count
            if self._norm_var:
                varss = self._stats[1, :-1] / count - means ** 2
                close_zero = np.isclose(varss, 0)
                if np.any(close_zero):
                    warnings.warn("0 variance encountered. Replacing with 1")
                    varss[close_zero] = 1
                scales = 1 / (varss ** 0.5)
            else:
                scales = 1
            vec *= scales
            vec -= means * scales
        else:
            if self._norm_var:
                raise ValueError(
                    "Unable to standardize the variance of a vector "
                    "with no global statistics"
                )
            else:
                warnings.warn("Standardizing a single vector to 0")
                vec[...] = 0
        return vec

    def _apply_tensor(self, tensor, axis, in_place):
        # apply transformation to tensor (with shape)
        num_coeffs = tensor.shape[axis]
        if self._stats is not None and self._stats.shape[1] != num_coeffs + 1:
            raise ValueError(
                "Expected feature vector of length {}; got {}".format(
                    self._stats.shape[1] - 1, num_coeffs
                )
            )
        other_axes = tuple(
            idx for idx in range(len(tensor.shape)) if idx != axis % len(tensor.shape)
        )
        if not in_place or tensor.dtype != np.float64:
            tensor = tensor.astype(np.float64)
        if self.have_stats:
            count = self._stats[0, -1]
            means = self._stats[0, :-1] / count
            varss = self._stats[1, :-1] / count - means ** 2
        elif sum(tensor.shape[idx] for idx in other_axes) == len(other_axes):
            if self._norm_var:
                raise ValueError(
                    "Unable to standardize the variance of a vector "
                    "with no global statistics"
                )
            else:
                warnings.warn("Standardizing a single vector to 0")
                tensor[...] = 0
                return tensor
        else:
            count = np.prod(tuple(tensor.shape[idx] for idx in other_axes))
            means = tensor.mean(axis=other_axes)
            varss = (tensor ** 2).sum(axis=other_axes) / count - means ** 2
        if self._norm_var:
            close_zero = np.isclose(varss, 0)
            if np.any(close_zero):
                warnings.warn("0 variance encountered. Replacing with 1")
                varss[close_zero] = 1
            scales = 1 / (varss ** 0.5)
        else:
            scales = np.ones(1)
        tensor_slice = [None] * len(tensor.shape)
        tensor_slice[axis] = slice(None)
        tensor_slice = tuple(tensor_slice)
        tensor *= scales[tensor_slice]
        tensor -= (means * scales)[tensor_slice]
        return tensor

    def apply(
        self, features: np.ndarray, axis: int = -1, in_place: bool = False
    ) -> np.ndarray:
        if (features.shape and not np.prod(features.shape)) or not len(features):
            raise ValueError("Cannot apply to empty array")
        if features.shape and features.ndim > 1:
            return self._apply_tensor(features, axis, in_place)
        else:
            return self._apply_vector(features, in_place)

    def save(
        self,
        wfilename: str,
        key: Optional[str] = None,
        compress: bool = False,
        overwrite: bool = True,
    ) -> None:
        r"""Save accumulated statistics to file

        If `wfilename` ends in :obj:`'.npy'`, stats will be written using
        :func:`numpy.save`.

        If `wfilename` ends in :obj:`'.npz'`, stats will be written to a numpy archive.
        If `overwrite` is :obj:`False`, other key-values will be loaded first if
        possible, then resaved. If `key` is set, data will be indexed by `key` in the
        archive. Otherwise, the data will be stored at the first unused key of the
        pattern :obj:`'arr_\d+'`. If `compress` is :obj:`True`,
        :func:`numpy.savez_compressed` will be used over :func:`numpy.savez`.

        Otherwise, data will be written using :func:`numpy.ndarray.tofile`

        Parameters
        ----------
        wfilename
        key
        compress
        overwrite

        Raises
        ------
        ValueError
            If no stats have been accumulated
        """
        if not self.have_stats:
            raise ValueError("No stats have been accumulated to save")
        if wfilename.endswith(".npy"):
            np.save(wfilename, self._stats)
        elif wfilename.endswith(".npz"):
            array = dict()
            if overwrite:
                try:
                    array = np.load(wfilename)
                except IOError:
                    pass
            if key is None:
                for key in ("arr_{}".format(v) for v in count(0)):
                    if key not in array:
                        break
            array[key] = self._stats
            if compress:
                np.savez_compressed(wfilename, **array)
            else:
                np.savez(wfilename, **array)
        else:
            self._stats.tofile(wfilename)


CMVN = Standardize


class Deltas(PostProcessor):
    r"""Calculate feature deltas (weighted rolling averages)

    Deltas are calculated by correlating the feature tensor with a 1D delta filter by
    enumerating over all but one axis (the "time axis" equivalent).

    Intermediate values are calculated with 64-bit floats, then cast back to the input
    data type.

    :class:`Deltas` will increase the size of the feature tensor when `num_deltas` is
    positive and passed features are non-empty.

    If `concatenate` is :obj:`True`, `target_axis` specifies the axis along which new
    deltas are appended. For example,

    >>> deltas = Deltas(num_deltas=2, concatenate=True, target_axis=1)
    >>> features_shape = list(features.shape)
    >>> features_shape[1] *= 3
    >>> assert deltas.apply(features).shape == tuple(features_shape)

    If `concatenate` is :obj:`False`, `target_axis` dictates the location of a new axis
    in the resulting feature tensor that will index the deltas (0 for the original
    features, 1 for deltas, 2 for double deltas, etc.). For example:

    >>> deltas = Deltas(num_deltas=2, concatenate=False, target_axis=1)
    >>> features_shape = list(features.shape)
    >>> features_shape.insert(1, 3)
    >>> assert deltas.apply(features).shape == tuple(features_shape)

    Deltas act as simple low-pass filters. Flipping the direction of the real filter to
    turn the delta operation into a simple convolution, the first order delta is defined
    as

    .. math::

         f(t) = \begin{cases}
            \frac{-t}{Z} & -W \leq t \leq W \\
            0 & \mathrm{else}
         \end{cases}

    where

    .. math:: Z = \sum_t f(t)^2

    for some :math:`W \geq 1`. Its Fourier Transform is

    .. math::

         F(\omega) = \frac{-2i}{Z\omega^2}\left(
            W\omega \cos W\omega - \sin W\omega \right)

    Note that it is completely imaginary. For :math:`W \geq 2`, :math:`F` is bound below
    :math:`\frac{i}{\omega}`. Hence, :math:`F` exhibits low-pass characteristics. Second
    order deltas are generating by convolving :math:`f(-t)`` with itself, third order is
    an additional :math:`f(-t)``, etc. By the convolution theorem, higher order deltas
    have Fourier responses that become tighter around :math:`F(0)`` (more lowpass).

    Parameters
    ----------
    num_deltas
    target_axis
    concatenate
    context_window
        The length of the filter to either side of the window. Positive.
    pad_mode
        How to pad the input sequence when correlating
    **kwargs
        Additional keyword arguments to be passed to :func:`numpy.pad`
    """

    aliases = {"deltas"}  #:
    concatenate: bool  #:
    num_deltas: int  #:

    def __init__(
        self,
        num_deltas: int,
        target_axis: int = -1,
        concatenate: bool = True,
        context_window: int = 2,
        pad_mode: Union[str, Callable] = "edge",
        **kwargs,
    ):
        self._target_axis = target_axis
        self._pad_mode = pad_mode
        self._pad_kwargs = kwargs
        self.concatenate = bool(concatenate)
        self.num_deltas = num_deltas
        self._filts = [np.ones(1, dtype=np.float64)]
        delta_filter = np.arange(1 + 2 * context_window, dtype=np.float64)
        delta_filter -= context_window
        delta_filter /= np.sum(delta_filter ** 2)
        for idx in range(num_deltas):
            self._filts.append(np.convolve(self._filts[idx], delta_filter))

    def apply(
        self, features: np.ndarray, axis: int = -1, in_place: bool = False
    ) -> np.ndarray:
        delta_feats = [features]
        other_axes = tuple(
            idx for idx in range(features.ndim) if idx != axis % features.ndim
        )
        other_shapes = tuple(features.shape[idx] for idx in other_axes)
        feat_slice = [slice(None)] * features.ndim
        for filt in self._filts[1:]:
            delta_feat = np.empty(features.shape, dtype=features.dtype)
            max_offset = (len(filt) - 1) // 2
            for other_indices in np.ndindex(other_shapes):
                for axis_idx, idx in zip(other_axes, other_indices):
                    feat_slice[axis_idx] = idx
                delta_feat[tuple(feat_slice)] = np.correlate(
                    np.pad(
                        features[tuple(feat_slice)].astype(np.float64, copy=False),
                        (max_offset, max_offset),
                        self._pad_mode,
                        **self._pad_kwargs,
                    ),
                    filt,
                    "full",
                )[len(filt) - 1 : -len(filt) + 1].astype(features.dtype, copy=False)
            delta_feats.append(delta_feat)
        if self.concatenate:
            return np.concatenate(delta_feats, self._target_axis)
        else:
            return np.stack(delta_feats, self._target_axis)


class Stack(PostProcessor):
    """Stack contiguous feature vectors together

    Parameters
    ----------
    num_vectors
        The number of subsequent feature vectors in time to be stacked.
    time_axis
        The axis along which subsequent feature vectors are drawn.
    pad_mode
        Specified how the axis in time will be padded on the right in order to be
        divisible by `num_vectors`. Additional keyword arguments will be passed to
        :func:`numpy.pad`. If unspecified, frames will instead be discarded in order to
        be divisible by `num_vectors`.
    """

    aliases = {"stack"}  #:
    num_vectors: int  #:
    time_axis: int  #:

    def __init__(
        self,
        num_vectors: int,
        time_axis: int = 0,
        pad_mode: Optional[Union[str, Callable]] = None,
        **kwargs,
    ) -> None:
        if num_vectors < 1:
            raise ValueError(f"Expected num_vectors to be positive, got {num_vectors}")
        self.num_vectors = num_vectors
        self.time_axis = time_axis
        self._pad_mode = pad_mode
        self._pad_kwargs = kwargs

    def apply(
        self, features: np.ndarray, axis: int = -1, in_place: bool = False
    ) -> np.ndarray:
        axis = axis % features.ndim
        time_axis = self.time_axis % features.ndim
        if axis == time_axis:
            raise RuntimeError(f"feature and time axes are the same ({axis})")
        shape = list(features.shape)
        T, F = shape[time_axis], shape[axis]
        if self._pad_mode is not None:
            rem = T % self.num_vectors
            if rem:
                padding = [(0, 0)] * features.ndim
                padding[time_axis] = (0, self.num_vectors - rem)
                features = np.pad(features, padding, self._pad_mode, **self._pad_kwargs)
                in_place = True
                T += self.num_vectors - rem
        nT, nF = T // self.num_vectors, F * self.num_vectors
        T = nT * self.num_vectors
        if features.ndim == 2:
            if not in_place:
                features = features.copy()
            if time_axis:
                features = features.T
            features = features[:T]
            features = features.reshape(nT, nF)
            if time_axis:
                features = features.T
        else:
            feat_slice = [slice(None)] * features.ndim
            buffs = []
            for i in range(self.num_vectors):
                feat_slice[time_axis] = slice(i, T, self.num_vectors)
                buffs.append(features[tuple(feat_slice)])
            features = np.concatenate(buffs, axis)
        return features



# Copyright 2022 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality to do with alias factories"""

import abc
from typing import Any, Mapping, Set, Type, TypeVar, Union, Type

__all__ = [
    "alias_factory_subclass_from_arg",
    "AliasedFactory",
]

T = TypeVar("T", bound="AliasedFactory", covariant=True)


class AliasedFactory(abc.ABC):
    """An abstract interface for initialing concrete subclasses with aliases"""

    aliases: Set[str] = set()
    """class aliases for :func:`from_alias`"""

    @classmethod
    def from_alias(cls: Type[T], alias: str, *args, **kwargs) -> T:
        """Factory method for initializing a subclass that goes by an alias

        All subclasses of this class have the class attribute `aliases`. This method
        matches `alias` to an element in some subclass' `aliases` and initializes it.
        Aliases of this class are included in the search. Alias conflicts are resolved
        by always trying to initialize the last registered subclass that matches the
        alias.

        Parameters
        ----------
        alias
            Alias of the subclass
        *args
            Positional arguments to initialize the subclass
        **kwargs
            Keyword arguments to initialize the subclass

        Raises
        ------
        ValueError
            Alias can't be found
        """
        stack = [cls]
        pushed_children = set()
        while stack:
            parent = stack.pop()
            if parent not in pushed_children:
                children = parent.__subclasses__()
                stack.append(parent)
                stack.extend(children)
                pushed_children.add(parent)
            elif alias in parent.aliases:
                return parent(*args, **kwargs)
        raise ValueError(f"Cannot find subclass with alias '{alias}'")


def alias_factory_subclass_from_arg(
    factory_class: Type[T], arg: Union[T, str, Mapping[str, Any]]
) -> T:
    """Boilerplate for getting an instance of an AliasedFactory

    Rather than an instance itself, a function could receive the arguments to initialize
    an :class:`AliasedFactory` with :func:`AliasedFactory.from_alias`. This function
    uses the following strategy to try and do so

    1. If `arg` is an instance of `factory_class`, return `arg`
    2. If `arg` is a :class:`str`, use it as the alias
    3. a. Copy `arg` to a dictionary
       b. Pop the key :obj:`'alias'` and treat the rest as keyword arguments
       c. If the key :obj:`'alias'` is not found, try :obj:`'name'`

    This function is intentionally limited in order to work nicely with JSON config
    files.
    """
    if isinstance(arg, factory_class):
        return arg
    elif isinstance(arg, str):
        return factory_class.from_alias(arg)
    else:
        arg = dict(arg)
        try:
            alias = arg.pop("alias")
        except KeyError:
            alias = arg.pop("name")
        return factory_class.from_alias(alias, **arg)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../../src"))


# -- Project information -----------------------------------------------------

project = "pydrobert-speech"
copyright = "2022, Sean Robertson"
author = "Sean Robertson"
language = "en"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinxcontrib.programoutput",
    "sphinx_autodoc_typehints",
]

naploeon_numpy_docstring = True

intersphinx_mapping = {
    "scipy": ("http://docs.scipy.org/doc/scipy/reference/", None),
    "matplotlib": ("https://matplotlib.org", None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "pydrobert.kaldi": ("https://pydrobert-kaldi.readthedocs.io/en/latest", None),
    "python": ("https://docs.python.org/", None),
    "soundfile": ("https://python-soundfile.readthedocs.io/en/latest/", None),
    "torch": ("https://pytorch.org/docs/stable/", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

napoleon_numpy_docstring = True
napoleon_google_docstring = False
napoleon_include_init_with_doc = True
autodoc_mock_imports = [
    "numpy",
    "matplotlib.axes",
    "matplotlib.colors",
    "matplotlib.figure",
    "matplotlib",
]
autodoc_typehints = "none"
autodoc_type_aliases = napoleon_type_aliases = {"np.ndarray": "numpy.ndarray"}
autodoc_inherit_docstrings = False
napoleon_preprocess_types = True
typehints_document_rtype = False
napoleon_use_rtype = False


# -- Options for HTML output -------------------------------------------------
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


highlight_language = "none"

master_doc = "index"



# Copyright 2021 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Scaling functions

Scaling functions transform a scalar in the frequency domain to some other real domain
(the "scale" domain). The scaling functions should be invertible. Their primary purpose
is to define the bandwidths of filters in :mod:`pydrobert.speech.filters`.
"""


import abc

import numpy as np

from pydrobert.speech.alias import AliasedFactory

__all__ = [
    "BarkScaling",
    "LinearScaling",
    "MelScaling",
    "OctaveScaling",
    "ScalingFunction",
]


class ScalingFunction(AliasedFactory):
    """Converts a frequency to some scale and back again"""

    @abc.abstractmethod
    def scale_to_hertz(self, scale: float) -> float:
        """Convert scale to frequency (in Hertz)"""
        pass

    @abc.abstractmethod
    def hertz_to_scale(self, hertz: float) -> float:
        """Convert frequency (in Hertz) to scalar"""
        pass


class LinearScaling(ScalingFunction):
    """Linear scaling between high and low scales/frequencies

    Parameters
    ----------
    low_hz
        The frequency (in Hertz) corresponding to scale 0.
    slope_hz
        The increase in scale corresponding to a 1 Hertz increase in frequency.
    """

    low_hz: float  #:
    slop_hz: float  #:
    aliases = {"linear", "uniform"}  #:

    def __init__(self, low_hz: float, slope_hz: float = 1.0):
        self.low_hz = low_hz
        self.slope_hz = slope_hz

    def scale_to_hertz(self, scale: float) -> float:
        return scale / self.slope_hz + self.low_hz

    def hertz_to_scale(self, hertz: float) -> float:
        return (hertz - self.low_hz) * self.slope_hz


class OctaveScaling(ScalingFunction):
    """Uniform scaling in log2 domain from low frequency

    Parameters
    ----------
    low_hz
        The positive frequency (in Hertz) corresponding to scale 0. Frequencies below
        this value should never be queried.
    """

    low_hz: float  #:
    aliases = {"octave"}  #:

    def __init__(self, low_hz: float):
        if low_hz <= 0:
            raise ValueError("low_hz must be positive")
        self.low_hz = low_hz

    def scale_to_hertz(self, scale: float) -> float:
        return (2 ** scale) * max(1e-10, self.low_hz)

    def hertz_to_scale(self, hertz: float) -> float:
        return np.log2(hertz / max(1e-10, self.low_hz))


class MelScaling(ScalingFunction):
    r"""Psychoacoustic scaling function

    Based of the experiment in [stevens1937]_ wherein participants adjusted a second
    tone until it was half the pitch of the first. The functional approximation to the
    scale is implemented with the formula from [oshaughnessy1987]_ (being honest, from
    `Wikipedia <https://en.wikipedia.org/wiki/Mel_scale>`__):

    .. math::

        s = 1127 \ln \left(1 + \frac{f}{700} \right)

    Where :math:`s` is the scale and :math:`f` is the frequency in Hertz.
    """

    aliases = {"mel"}  #:

    def scale_to_hertz(self, scale: float) -> float:
        return 700.0 * (np.exp(scale / 1127.0) - 1.0)

    def hertz_to_scale(self, hertz: float) -> float:
        return 1127.0 * np.log(1 + hertz / 700.0)


class BarkScaling(ScalingFunction):
    r"""Psychoacoustic scaling function

    Based on a collection experiments briefly mentioned in [zwicker1961]_ involving
    masking to determine critical bands. The functional approximation to the scale is
    implemented with the formula from [traunmuller1990]_ (being honest, from `Wikipedia
    <https://en.wikipedia.org/wiki/Bark_scale>`__):

    .. math::

         s = \begin{cases}
            z + 0.15(2 - z) & \mbox{if }z < 2 \\
            z + 0.22(z - 20.1) & \mbox{if }z > 20.1
         \end{cases}

    where

    .. math::

        z = 26.81f/(1960 + f) - 0.53

    Where :math:`s` is the scale and :math:`f` is the frequency in Hertz.
    """

    aliases = {"bark"}  #:

    def scale_to_hertz(self, scale: float) -> float:
        bark = None
        if scale < 2:
            bark = (20.0 * scale - 6.0) / 17.0
        elif scale > 20.1:
            bark = (50.0 * scale + 221.1) / 61.0
        else:
            bark = scale
        return 1960.0 * (bark + 0.53) / (26.28 - bark)

    def hertz_to_scale(self, hertz: float) -> float:
        bark = 26.81 * hertz / (1960.0 + hertz) - 0.53
        if bark < 2:
            return bark + 0.15 * (2.0 - bark)
        elif bark > 20.1:
            return bark + 0.22 * (bark - 20.1)
        else:
            return bark



# Copyright 2021 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Miscellaneous utility functions"""

import warnings

from re import match
from typing import Any, Optional

import pydrobert.speech.config as config
import numpy as np

from pydrobert.speech.alias import alias_factory_subclass_from_arg as _afsfa


__all__ = [
    "angular_to_hertz",
    "circshift_fourier",
    "gauss_quant",
    "hertz_to_angular",
    "read_signal",
]


def alias_factory_subclass_from_arg(*args, **kwargs):
    warnings.warn(
        "using alias_factory_subclass_from_arg from util is deprecated. "
        "Use from pydrobert.speech.alias instead",
        category=DeprecationWarning,
        stacklevel=2,
    )

    return _afsfa(*args, **kwargs)


def _gauss_quant_odeh_evans(p: float, mu: float = 0, std: float = 1) -> float:
    r = 1 - p if p > 0.5 else p
    if r < 1e-20:
        z = 10
    else:
        y = (-2 * np.log(r)) ** 0.5
        z = (
            ((4.53642210148e-5 * y + 0.0204231210245) * y + 0.342242088547) * y + 1
        ) * y + 0.322232431088
        z /= (
            ((0.0038560700634 * y + 0.10353775285) * y + 0.531103462366) * y
            + 0.588581570495
        ) * y + 0.099348462606
        z = y - z
    if p < 0.5:
        z = -z
    return z * std + mu


try:
    from scipy.stats import norm

    def gauss_quant(p: float, mu: float = 0, std: float = 1) -> float:
        return norm.ppf(p) * std + mu


except ImportError:
    gauss_quant = _gauss_quant_odeh_evans
gauss_quant.__doc__ = """\
Gaussian quantile function

Given a probability from a univariate Gaussian, determine the value of the random
variable such that the probability of drawing a value l.t.e. to that value is equal to
the probability. In other words, the so-called inverse cumulative distribution function.

If `scipy` can be imported, this function uses :func:`scipy.norm.ppf` to calculate the
result. Otherwise, it uses the approximation from Odeh & Evans 1974 (thru Brophy 1985)

Parameters
----------
p
    The probability
mu
    The Gaussian mean
std
    The Gaussian standard deviation

Returns
-------
q : float
    The random variable value
"""


def hertz_to_angular(hertz: float, samp_rate: float) -> float:
    """Convert cycles/sec to radians/sec"""
    return hertz * 2 * np.pi / samp_rate


def angular_to_hertz(angle: float, samp_rate: float) -> float:
    """Convert radians/sec to cycles/sec"""
    return angle * samp_rate / (2 * np.pi)


def circshift_fourier(
    filt: np.ndarray,
    shift: float,
    start_idx: int = 0,
    dft_size: int = None,
    copy: bool = True,
) -> np.ndarray:
    r"""Circularly shift a filter in the time domain, from the fourier domain

    A simple application of the shift theorem

    .. math::

        DFT(T_u x)[k] = DFT(x)[k] e^{-2i\pi k u}

    Where we set ``u = shift / dft_size``

    Parameters
    ----------
    filt
        The filter, in the fourier domain
    shift
        The number of samples to be translated by.
    start_idx
        If `filt` is a truncated frequency response, this parameter indicates at what
        index in the dft the nonzero region starts
    dft_size
        The dft_size of the filter. Defaults to
        ``len(filt) + start_idx``
    copy
        Whether it is okay to modify and return `filt`

    Returns
    -------
    out : np.ndarray
        The 128-bit complex filter frequency response, shifted by `u`
    """
    shift %= dft_size
    if dft_size is None:
        dft_size = len(filt) + start_idx
    if copy or filt.dtype != np.complex128:
        return filt * np.exp(
            -2j
            * np.pi
            * shift
            / dft_size
            * (np.arange(start_idx, start_idx + len(filt),) % dft_size)
        )
    else:
        filt *= np.exp(
            -2j
            * np.pi
            * shift
            / dft_size
            * (np.arange(start_idx, start_idx + len(filt),) % dft_size)
        )
        return filt


def _kaldi_table_read_signal(rfilename, dtype, key, **kwargs):
    from pydrobert.kaldi.io import open as io_open  # type: ignore

    if key is None:
        key = 0
    if dtype is None:
        dtype = "bm"
    if isinstance(key, str):
        with io_open(rfilename, dtype, mode="r+", **kwargs) as table:
            return table[key]
    else:
        with io_open(rfilename, dtype, mode="r", **kwargs) as table:
            for idx in range(key):
                if not table.move():
                    raise IndexError("table index out of range")
            return table.value()


def _scipy_io_read_signal(rfilename, dtype, key, **kwargs):
    from scipy.io import wavfile  # type: ignore

    _, data = wavfile.read(rfilename, **kwargs)
    if dtype:
        data = data.astype(dtype)
    return data


def _wave_read_signal(rfilename, dtype, key, **kwargs):
    import wave

    wave_file = wave.open(rfilename, **kwargs)
    try:
        dtype_in = "<i{}".format(wave_file.getsampwidth())
        data = np.frombuffer(
            wave_file.readframes(wave_file.getnframes()), dtype=dtype_in
        )
        n_data_points = len(data)
        n_channels = wave_file.getnchannels()
        if n_data_points % n_channels:
            raise IOError("Number of channels do not evenly divide wave samples")
        if n_channels > 1:
            data = data.reshape((n_data_points // n_channels, n_channels), order="C")
    finally:
        wave_file.close()
    if dtype:
        data = data.astype(dtype)
    return data


def _hdf5_read_signal(rfilename, dtype, key, **kwargs):
    import h5py  # type: ignore

    with h5py.File(rfilename, "r", **kwargs) as h5py_file:
        if key:
            data = h5py_file[key]
        else:
            group_stack = [h5py_file]
            data = None
            while group_stack:
                cur_group = group_stack.pop()
                if isinstance(cur_group, h5py.Dataset):
                    data = cur_group
                    break
                else:
                    keys = list(cur_group.keys())
                    keys.sort(reverse=True)
                    for name in keys:
                        group_stack.append(cur_group[name])
            if data is None:
                raise IOError("Could not find any dataset")
        if dtype:
            data = np.array(data, dtype=dtype)
        else:
            data = np.array(data)
    return data


def _numpy_binary_read_signal(rfilename, dtype, key, **kwargs):
    data = np.load(rfilename, **kwargs)
    if dtype:
        data = data.astype(dtype)
    return data


def _numpy_archive_read_signal(rfilename, dtype, key, **kwargs):
    archive = np.load(rfilename, **kwargs)
    if key:
        data = archive[key]
    else:
        data = archive["arr_0"]
    if dtype:
        data = data.astype(dtype)
    return data


def _torch_read_signal(rfilename, dtype, key, **kwargs):
    import torch  # type: ignore

    data = torch.load(rfilename, map_location="cpu", **kwargs).numpy()
    if dtype:
        data = data.astype(dtype)
    return data


def _kaldi_input_read_signal(rfilename, dtype, key, **kwargs):
    from pydrobert.kaldi.io import open as io_open  # type: ignore

    if dtype is None:
        dtype = "bm"
    with io_open(rfilename, mode="r", **kwargs) as inp_stream:
        data = inp_stream.read(dtype)
    return data


def _numpy_fromfile_read_signal(rfilename, dtype, key, **kwargs):
    if dtype:
        data = np.fromfile(rfilename, dtype=dtype, **kwargs)
    else:
        data = np.fromfile(rfilename, **kwargs)
    return data


def _soundfile_read_signal(rfilename, dtype, key, **kwargs):
    import soundfile

    fi = soundfile.info(rfilename)
    if fi.subtype == "FLOAT":
        dtype_ = np.float32
    elif fi.subtype == "DOUBLE":
        dtype_ = np.float64
    elif fi.subtype == "PCM_S8":
        dtype_ = np.int8
    elif fi.subtype == {"PCM_U8"}:
        dtype_ = np.uint8
    elif fi.subtype in {"PCM_32", "PCM_24"}:
        dtype_ = np.int32
    else:
        # FIXME(sdrobert): PCM_16 is a decent guess for the remainder of types, but
        # it's definitely not complete
        dtype_ = np.int16
    data = soundfile.read(rfilename, dtype=dtype_, **kwargs)[0]
    if dtype is not None:
        # if you don't do this as a second stage and you want floats out the back,
        # soundfile will scale those to the range +/- 1. Other decoders are two-stage
        # as well.
        data = data.astype(dtype)
    return data


def read_signal(
    rfilename: str,
    dtype: Optional[np.dtype] = None,
    key: Any = None,
    force_as: Optional[str] = None,
    **kwargs,
) -> np.ndarray:
    r"""Read a signal from a variety of possible sources

    Though the goal of this function is to return an array representing a signal of some
    sort, the way it goes about doing so depends on the setting of `rfilename`,
    processed in the following order:

    1.  If `rfilename` starts with the regular expression :obj:`r'^(ark|scp)(,\w+)*:'`,
        the file is treated as a Kaldi table and opened with the kaldi data type `dtype`
        (defaults to :class:`BaseMatrix`). The package :mod:`pydrobert.kaldi` will be
        imported to handle reading. If `key` is set, the value associated with that key
        is retrieved. Otherwise the first listed value is returned.
    2.  If `rfilename` ends with a file type listed in
        :obj:`pydrobert.speech.config.SOUNDFILE_SUPPORTED_FILE_TYPES` (requires
        :mod:`soundfile`), the file will be opened with that audio file type.
    3.  If `rfilename` ends with :obj:`'.wav'`, the file is assumed to be a wave file.
        The function will rely on the :mod:`scipy` package to load the file if
        :mod:`scipy` can be imported. Otherwise, it uses the standard :mod:`wave`
        package. The type of data encodings each package can handle varies, though
        neither can handle compressed data.
    4.  If `rfilename` ends with :obj:`'.hdf5'`, the file is assumed to be an HDF5 file.
        HDF5 and :mod:`h5py` must be installed on the host system to read this way. If
        `key` is set, the data will assumed to be indexed by `key` on the archive.
        Otherwise, a depth-first search of the archive will be performed for the first
        data set. If set, data will be cast to as the numpy data type `dtype`
    5.  If `rfilename` ends with :obj:`'.npy'`, the file is assumed to be a binary in
        Numpy format. If set, the result will be cast as the numpy data type `dtype`.
    6.  If `rfilename` ends with :obj:`'.npz'`, the file is assumed to be an archive in
        Numpy format. If `key` is swet, the data indexed by `key` will be loaded.
        Otherwise the data indexed by the key :obj:`'arr_0'` will be loaded. If set, the
        result will be cast as the numpy data type `dtype`.
    7.  If `rfilename` ends with :obj:`'.pt'`, the file is assumed to be a binary in
        PyTorch format. If set, the results will be cast as the numpy data type `dtype`.
    8.  If `rfilename` ends with :obj:`'.sph'`, the file is assumed to be a NIST SPHERE
        file. If set, the results will be cast as the numpy data type `dtype`
    9.  If `rfilename`` ends with ``'|'``, it will try to read an object of kaldi data
        type `dtype` (defaults to :class:`BaseMatrix`) from a basic kaldi input stream.
    10. Otherwise, we throw an :class:`IOError`

    Additional keyword arguments are passed along to the associated
    open or read operation.

    Parameters
    ----------
    rfilename 
    dtype
    key
    force_as
        If not :obj:`None`, forces `rfilename` to be interpreted as a specific
        file type, bypassing the above selection strategy. ``'tab'``: Kaldi
        table; ``'wav'``: wave file; ``'hdf5'``: HDF5 file; ``'npy'``: Numpy
        binary; ``'npz'``: Numpy archive; ``'pt'``: PyTorch binary; ``'sph'``:
        NIST sphere; ``'kaldi'`` Kaldi object; ``'file'`` read via
        :func:`numpy.fromfile`. The types in :obj:`SOUNDFILE_SUPPORTED_FILE_TYPES`
        are also valid values. `'soundfile'` will use :mod:`soundfile` to read the file
        regardless of the suffix.
    **kwargs

    Returns
    -------
    signal : np.ndarray

    Warnings
    --------
    Post v 0.2.0, the behaviour after step 8 changed. Instead of trying to read first as
    Kaldi input, and, failing that, via :func:`numpy.fromfile`, we try to read as Kaldi
    input if the file name ends with ``'|'`` and error otherwise. The catch-all
    behaviour was disabled due to the interaction with
    :obj:`pydrobert.speech.config.SOUNDFILE_SUPPORTED_FILE_TYPES` whose value depends on
    the existence of :mod:`soundfile` and the underlying version of `libsndfile
    <https://libsndfile.github.io/libsndfile>`__.

    Notes
    -----
    Python code for reading SPHERE files (not via :mod:soundfile`) was based off of
    `sph2pipe v 2.5
    <https://www.ldc.upenn.edu/language-resources/tools/sphere-conversion-tools>`__.
    That code can only suppport the "shorten" audio format up to version 2.
    """
    if force_as is None:
        if match(r"^(ark|scp)(,\w+)*:", rfilename):
            force_as = "table"
        elif (
            rfilename.rsplit(".", maxsplit=1)[-1]
            in config.SOUNDFILE_SUPPORTED_FILE_TYPES
        ):
            force_as = "soundfile"
        elif rfilename.endswith(".wav"):
            force_as = "wav"
        elif rfilename.endswith(".hdf5"):
            force_as = "hdf5"
        elif rfilename.endswith(".npy"):
            force_as = "npy"
        elif rfilename.endswith(".npz"):
            force_as = "npz"
        elif rfilename.endswith(".pt"):
            force_as = "pt"
        elif rfilename.endswith(".sph"):
            force_as = "sph"
        elif rfilename.endswith("|"):
            force_as = "kaldi"
        else:
            raise IOError(f"Unable to infer file type from {rfilename}. Set force_as.")
    if force_as == "table":
        data = _kaldi_table_read_signal(rfilename, dtype, key, **kwargs)
    elif force_as == "wav":
        try:
            data = _scipy_io_read_signal(rfilename, dtype, key, **kwargs)
        except ImportError:
            data = _wave_read_signal(rfilename, dtype, key, **kwargs)
    elif force_as == "hdf5":
        data = _hdf5_read_signal(rfilename, dtype, key, **kwargs)
    elif force_as == "npy":
        data = _numpy_binary_read_signal(rfilename, dtype, key, **kwargs)
    elif force_as == "npz":
        data = _numpy_archive_read_signal(rfilename, dtype, key, **kwargs)
    elif force_as == "pt":
        data = _torch_read_signal(rfilename, dtype, key, **kwargs)
    elif force_as == "sph":
        from ._sphere import sphere_read_signal

        data = sphere_read_signal(rfilename, dtype, key, **kwargs)
    elif force_as == "kaldi":
        data = _kaldi_input_read_signal(rfilename, dtype, key, **kwargs)
    elif force_as == "file":
        data = _numpy_fromfile_read_signal(rfilename, dtype, key, **kwargs)
    elif force_as == "soundfile":
        data = _soundfile_read_signal(rfilename, dtype, key, **kwargs)
    else:
        avail_force_as = {
            "table",
            "wav",
            "hdf5",
            "npy",
            "npz",
            "pt",
            "sph",
            "kaldi",
            "file",
            "soundfile",
        } | config.SOUNDFILE_SUPPORTED_FILE_TYPES
        msg = f"force_as ('{force_as}') is not one of {avail_force_as}."
        if force_as in config._BASE_SOUNDFILE_SUPPORTED_TYPES:
            msg += (
                "\n... but it could be, with the proper version of libsndfile and "
                "pysoundfile installed"
            )
        elif force_as in config._FULL_SOUNDFILE_SUPPORTED_TYPES:
            msg += (
                "\n... but pysoundfile may be able to handle it. "
                "Try setting force_as = 'soundfile'"
            )
        raise ValueError(msg)
    return data




# pylint: skip-file


import os

from math import erf

import numpy as np
import pytest

from pydrobert.speech import util


@pytest.mark.parametrize("shift", [0, 1, 100, -100])
@pytest.mark.parametrize("dft_size", [1, 2, 51, 1000], ids=["l1", "l2", "l51", "l1000"])
@pytest.mark.parametrize("copy", [True, False], ids=["copy", "keep"])
@pytest.mark.parametrize("start_idx", [0, 1, -1], ids=["at", "after", "before"])
def test_circshift_fourier(shift, dft_size, start_idx, copy):
    start_idx %= dft_size
    zeros = np.random.randint(dft_size)
    X = 10 * np.random.random(dft_size - zeros) + 10j * np.random.random(
        dft_size - zeros
    )
    Xs = util.circshift_fourier(
        X.copy(), shift, start_idx=start_idx, dft_size=dft_size, copy=copy
    )
    X = np.roll(np.pad(X, (0, zeros), "constant"), start_idx)
    Xs = np.roll(np.pad(Xs, (0, zeros), mode="constant"), start_idx)
    assert len(X) == len(Xs)
    x = np.fft.ifft(X)
    xs = np.fft.ifft(Xs)
    assert np.allclose(np.roll(x, shift), xs)


@pytest.mark.parametrize("mu", [0, -1, 100])
@pytest.mark.parametrize("std", [0.1, 1, 10])
@pytest.mark.parametrize("do_scipy", [True, False])
def test_gauss_quant(mu, std, do_scipy):
    X = np.arange(1000, dtype=float) / 1000 - 0.5
    X /= X.std()
    X *= std / 2
    X += mu
    for x in X:
        p = 0.5 * (1 + erf((x - mu) / std / np.sqrt(2)))
        if do_scipy:
            pytest.importorskip("scipy.norm")
            x2 = util.gauss_quant(p, mu=mu, std=std)
        else:
            # because we don't give access to this if scipy is
            # installed, we have to access the private function
            x2 = util._gauss_quant_odeh_evans(p, mu=mu, std=std)
        assert np.isclose(x, x2, atol=1e-5)


@pytest.mark.parametrize("key", [True, False])
def test_read_table(temp_dir, key):
    kaldi = pytest.importorskip("pydrobert.kaldi.io")
    rxfilename = "ark:{}".format(os.path.join(temp_dir, "foo.ark"))
    key_1 = "lions"
    key_2 = "tigers"
    buff_1 = np.random.random((100, 10))
    buff_2 = np.random.random((1000, 2))
    with kaldi.open(rxfilename, "dm", "w") as table:
        table.write(key_1, buff_1)
        table.write(key_2, buff_2)
    if key:
        buff_3 = util.read_signal(rxfilename, dtype="dm", key=key_2)
        assert np.allclose(buff_2, buff_3)
    else:
        buff_3 = util.read_signal(rxfilename, dtype="dm")
        assert np.allclose(buff_1, buff_3)


@pytest.mark.parametrize("backend", ["wav", "scipy", "soundfile"])
@pytest.mark.parametrize("channels", [1, 2], ids=["mono", "stereo"])
@pytest.mark.parametrize("sampwidth", [2, 4])
def test_read_wave(temp_dir, backend, channels, sampwidth):
    import wave

    rfilename = os.path.join(temp_dir, "foo.wav")
    if channels > 1:
        wave_buffer_1 = np.random.random((1000, channels)) * 1000
    else:
        wave_buffer_1 = np.random.random(1000) * 1000
    wave_buffer_1 = wave_buffer_1.astype("<i{}".format(sampwidth))
    wave_bytes = wave_buffer_1.tobytes("C")
    wave_file = wave.open(rfilename, "wb")
    wave_file.setnchannels(channels)
    wave_file.setsampwidth(sampwidth)
    wave_file.setframerate(8000)
    wave_file.writeframes(wave_bytes)
    wave_file.close()
    if backend == "scipy":
        pytest.importorskip("scipy")
        wave_buffer_2 = util._scipy_io_read_signal(rfilename, None, None)
    elif backend == "soundfile":
        sf = pytest.importorskip("soundfile")
        if "WAV" not in sf.available_formats():
            pytest.skip("libsndfile cannot handle wav files")
        wave_buffer_2 = util._soundfile_read_signal(rfilename, None, None)
    else:
        wave_buffer_2 = util._wave_read_signal(rfilename, None, None)
    assert np.allclose(wave_buffer_1, wave_buffer_2)


@pytest.mark.parametrize(
    "name",
    [
        "123_1pcbe",
        "123_1pcle",
        "123_1ulaw",
        "123_2alaw",
        "123_2pcbe",
        "123_2pcle",
        "123_2ulaw",
    ],
)
def test_read_sphere(name):
    audio_dir = os.path.join(os.path.dirname(__file__), "audio")
    if name.endswith("alaw"):
        sph_file = os.path.join(audio_dir, name + ".sph")
    else:
        sph_file = os.path.join(audio_dir, name + "_shn.sph")
    wav_file = os.path.join(audio_dir, name + ".wav")
    assert os.path.isfile(sph_file)
    assert os.path.isfile(wav_file)
    wav = util.read_signal(wav_file)
    sph = util.read_signal(sph_file)
    assert np.all(sph == wav)


@pytest.mark.parametrize(
    "env_var,suffix", [("WSJ_DIR", ".wv1"), ("TIMIT_DIR", ".sph")], ids=["wsj", "timit"]
)
def test_read_sphere_corpus(temp_dir, env_var, suffix):
    num_utts = 50
    env_dir = os.environ.get(env_var, None)
    if env_dir is None:
        pytest.skip(f"Corpus dir not set (export {env_var})")
    sph2pipe_path = os.environ.get("SPH2PIPE", None)
    if sph2pipe_path is None:
        pytest.skip("SPH2PIPE dir not set")
    try:
        import pathlib
    except ImportError:
        pathlib = pytest.importorskip("pathlib2")
    sphere_files = [
        str(x)
        for (x, _) in zip(
            pathlib.Path(env_dir).glob("**/*" + suffix), list(range(num_utts))
        )
    ]
    import subprocess

    wav_files = []
    for utt_idx, src in enumerate(sphere_files):
        wav_file = os.path.join(temp_dir, "{}.wav".format(utt_idx))
        wav_files.append(wav_file)
        assert not subprocess.call([sph2pipe_path, "-f", "wav", src, wav_file])
    for wav_path, sph_path in zip(wav_files, sphere_files):
        wav = util.read_signal(wav_path)
        sph = util.read_signal(sph_path, force_as="sph")
        assert (wav == sph).all()


@pytest.mark.parametrize("key", [True, False])
def test_read_hdf5(temp_dir, key):
    h5py = pytest.importorskip("h5py")
    rfilename = os.path.join(temp_dir, "foo.hdf5")
    h5py_file = h5py.File(rfilename, "w")
    h5py_file.create_group("a/b/c")
    h5py_file.create_group("a/b/d/e")
    dset_1 = np.random.random((1000, 2000))
    dset_2 = (np.random.random(10) * 1000).astype(int)
    h5py_file.create_dataset("a/b/d/f", (1000, 2000), data=dset_1)
    h5py_file.create_dataset("g", (10,), data=dset_2)
    h5py_file.close()
    if key:
        dset_3 = util.read_signal(rfilename, key="g")
        assert np.allclose(dset_2, dset_3)
    else:
        dset_3 = util.read_signal(rfilename)
        assert np.allclose(dset_1, dset_3)


def test_read_torch(temp_dir):
    torch = pytest.importorskip("torch")
    torch.manual_seed(10)
    rfilename = os.path.join(temp_dir, "foo.pt")
    exp = torch.randn(10, 4)
    torch.save(exp, rfilename)
    exp = exp.numpy()
    act = util.read_signal(rfilename)
    assert np.allclose(exp, act)


@pytest.mark.parametrize(
    "allow_pickle", [True, False], ids=["picklable", "notpicklable"]
)
@pytest.mark.parametrize("fix_imports", [True, False], ids=["fix", "nofix"])
def test_read_numpy_binary(temp_dir, allow_pickle, fix_imports):
    rfilename = os.path.join(temp_dir, "foo.npy")
    buff_1 = np.random.random((1000, 10, 5))
    np.save(rfilename, buff_1, allow_pickle=allow_pickle, fix_imports=fix_imports)
    buff_2 = util.read_signal(rfilename)
    assert np.allclose(buff_1, buff_2)


@pytest.mark.parametrize(
    "compressed", [True, False], ids=["compressed", "uncompressed"]
)
@pytest.mark.parametrize("key", [True, False], ids=["withkey", "withoutkey"])
def test_read_numpy_archive(temp_dir, compressed, key):
    rfilename = os.path.join(temp_dir, "foo.npz")
    buff_1 = np.random.random((5, 1, 2))
    buff_2 = np.random.random((1,))
    if compressed and key:
        np.savez_compressed(rfilename, a=buff_1, b=buff_2)
    elif compressed:
        np.savez_compressed(rfilename, buff_1, buff_2)
    elif key:
        np.savez(rfilename, a=buff_1, b=buff_2)
    else:
        np.savez(rfilename, buff_1, buff_2)
    if key:
        buff_3 = util.read_signal(rfilename, key="a")
    else:
        buff_3 = util.read_signal(rfilename)
    assert np.allclose(buff_1, buff_3)


@pytest.mark.parametrize("text", [True, False])
def test_read_numpy_fromfile(temp_dir, text):
    rfilename = os.path.join(temp_dir, "foo")
    buff_1 = np.random.random(1000)
    sep = "," if text else ""
    buff_1.tofile(rfilename, sep=sep)
    buff_2 = util.read_signal(rfilename, sep=sep, force_as="file")
    assert np.allclose(buff_1, buff_2)


@pytest.mark.parametrize("filetype", ["wav", "aiff", "flac", "ogg"])
def test_read_soundfile(filetype):
    sf = pytest.importorskip("soundfile")
    if filetype.upper() not in sf.available_formats():
        pytest.skip(f"This version of libsndfile does not support {filetype}")
    filename = os.path.join(os.path.dirname(__file__), "audio", "sin1k." + filetype)
    buff = util.read_signal(filename, dtype=np.float64)
    buff *= np.blackman(len(buff))
    pow = np.abs(np.fft.rfft(buff))
    # 8kHz/sec * 1 sec = 8000 samples
    # sin at 1kHz = sample 1000
    assert np.argmax(pow) == 1000




# Copyright 2021 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Speech processing library"""

import warnings

from .alias import AliasedFactory as _AF

__author__ = "Sean Robertson"
__email__ = "sdrobert@cs.toronto.edu"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2021 Sean Robertson"

__all__ = [
    "alias",
    "compute",
    "config",
    "corpus",
    "filters",
    "post",
    "pre",
    "scales",
    "util",
]


class AliasedFactory(_AF):
    @classmethod
    def from_alias(cls, alias: str, *args, **kwargs):
        warnings.warn(
            "using AliasedFactory from pydrobert.speech is deprecated. Import from "
            "pydrobert.speech.alias",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return super().from_alias(alias, *args, **kwargs)


try:
    from ._version import version as __version__  # type: ignore
except ImportError:
    __version__ = "inplace"




import os
import wave

import numpy as np
import pytest

from pydrobert.speech import command_line


def test_compute_feats_from_kaldi_tables(temp_dir):
    kaldi_io = pytest.importorskip("pydrobert.kaldi.io")
    np.random.seed(5)
    wav_scp = os.path.join(temp_dir, "wav.scp")
    feat_ark = os.path.join(temp_dir, "feat.ark")
    wav_file_dir = os.path.join(temp_dir, "wav")
    num_utts = 100
    max_samples = 32000
    # sampling_rate = 16000
    num_digits = int(np.log10(num_utts)) + 1
    utt_fmt = "utt{{:0{}d}}".format(num_digits)
    computer_json_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "data", "fbank.json"
    )
    preprocessor_json_path = os.path.join(temp_dir, "preprocess.json")
    postprocessor_json_path = os.path.join(temp_dir, "unit.json")
    with open(preprocessor_json_path, "w") as f:
        f.write('["dither"]\n')
    with open(postprocessor_json_path, "w") as f:
        f.write('["unit"]\n')
    if not os.path.isdir(wav_file_dir):
        os.makedirs(wav_file_dir)
    with open(wav_scp, "w") as scp:
        for utt_idx in range(num_utts):
            utt_id = utt_fmt.format(utt_idx)
            wav_path = os.path.join(wav_file_dir, utt_id + ".wav")
            num_samples = np.random.randint(max_samples)
            signal = np.random.randint(
                -(2 ** 15), 2 ** 15 - 1, num_samples, dtype=np.int16
            )
            wv = wave.open(wav_path, "wb")
            wv.setnchannels(1)
            wv.setsampwidth(2)
            wv.setframerate(16000)
            wv.writeframes(signal.tobytes())
            wv.close()
            scp.write("{} {}\n".format(utt_id, wav_path))
    args = [
        "scp,s:" + wav_scp,
        "ark:" + feat_ark,
        computer_json_path,
        "--seed=30",
        "--preprocess={}".format(preprocessor_json_path),
        "--postprocess={}".format(postprocessor_json_path),
    ]
    assert not command_line.compute_feats_from_kaldi_tables(args)
    exps = []
    with kaldi_io.open("ark,s:" + feat_ark, "bm") as feats:
        not_all_feats = 0
        for utt_idx, feat in enumerate(feats):
            not_all_feats |= feat.shape[0]
            assert feat.shape[0] == 0 or feat.shape[-1] == 40
            exps.append(feat)
        assert utt_idx == num_utts - 1
        assert not_all_feats
    np.random.seed(300)
    assert not command_line.compute_feats_from_kaldi_tables(args)
    with kaldi_io.open("ark,s:" + feat_ark, "bm") as feats:
        for utt_idx, (exp, act) in enumerate(zip(exps, feats)):
            assert np.allclose(act, exp)


def test_signals_to_torch_feat_dir(temp_dir):
    torch = pytest.importorskip("torch")
    torch.manual_seed(50)
    feat_dir = os.path.join(temp_dir, "feat")
    raw_dir = os.path.join(temp_dir, "raw")
    map_path = os.path.join(temp_dir, "map")
    computer_json_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "data", "fbank.json"
    )
    preprocessor_json_path = os.path.join(temp_dir, "preprocess.json")
    with open(preprocessor_json_path, "w") as f:
        f.write('["dither"]\n')
    assert os.path.isfile(computer_json_path)
    num_utts, max_samples = 100, 1600
    num_digits = torch.log10(torch.tensor(float(num_utts))).long().item() + 1
    utt_fmt = "utt{{:0{}d}}".format(num_digits)
    if not os.path.isdir(raw_dir):
        os.makedirs(raw_dir)
    utt_ids = []
    utt2signal = dict()
    with open(map_path, "w") as mp:
        for utt_idx in range(num_utts):
            utt_id = utt_fmt.format(utt_idx)
            utt_ids.append(utt_id)
            num_samples = torch.randint(1, max_samples, (1,)).long().item()
            signal = torch.randint(
                -(2 ** 15), 2 ** 15 - 1, (num_samples,), dtype=torch.float32
            )
            utt2signal[utt_id] = signal
            file_type = torch.randint(3, (1,)).long().item()
            if file_type == 2:  # wave file
                path = os.path.join(raw_dir, "{}.wav".format(utt_idx))
                wv = wave.open(path, "wb")
                wv.setnchannels(1)
                wv.setsampwidth(2)
                wv.setframerate(16000)
                wv.writeframes(signal.to(torch.int16).numpy().tobytes())
                wv.close()  # py2.7 doesn't have context manager
            elif file_type == 1:  # npy file
                path = os.path.join(raw_dir, "{}.npy".format(utt_idx))
                np.save(path, signal.numpy())
            else:  # torch file
                path = os.path.join(raw_dir, "{}.pt".format(utt_idx))
                torch.save(signal, path)
            mp.write("{} {}\n".format(utt_id, path))
    args = [map_path, computer_json_path, feat_dir]
    assert not command_line.signals_to_torch_feat_dir(args)
    for utt_id in utt_ids:
        feat = torch.load(os.path.join(feat_dir, "{}.pt".format(utt_id)))
        assert feat.shape[-1] == 40
        del feat
    args.pop(1)  # remove computer -- storing raw
    torch.manual_seed(30)
    assert not command_line.signals_to_torch_feat_dir(args)
    for utt_id, exp in list(utt2signal.items()):
        act = torch.load(os.path.join(feat_dir, "{}.pt".format(utt_id)))
        assert torch.allclose(exp, act.flatten())
    torch.manual_seed(40)
    args.append("--seed=1")
    args.append("--preprocess={}".format(preprocessor_json_path))
    assert not command_line.signals_to_torch_feat_dir(args)
    for utt_id in utt_ids:
        signal = utt2signal[utt_id]
        noisy = torch.load(os.path.join(feat_dir, "{}.pt".format(utt_id)))
        assert not torch.allclose(signal, noisy.flatten())
        utt2signal[utt_id] = noisy
    torch.manual_seed(70)
    args.append("--num-workers=2")
    assert not command_line.signals_to_torch_feat_dir(args)
    for utt_id, exp in list(utt2signal.items()):
        act = torch.load(os.path.join(feat_dir, "{}.pt".format(utt_id)))
        assert torch.allclose(exp, act)



# Copyright 2021 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes for pre-processing speech signals"""


import abc
from typing import Optional

import numpy as np

from pydrobert.speech.alias import AliasedFactory

__all__ = [
    "PreProcessor",
    "Dither",
    "Preemphasize",
]


class PreProcessor(AliasedFactory):
    """A container for pre-processing signals with a transform"""

    @abc.abstractmethod
    def apply(self, signal: np.ndarray, axis: int = -1, in_place=False) -> np.ndarray:
        """Applies the transformation to a signal tensor

        Consult the class documentation for more details on what the transformation is.

        Parameters
        ----------
        signal
        axis
            The axis of `signal` to apply the transformation along
        in_place
            Whether it is okay to modify `signal` (:obj:`True`) or whether a copy should
            be made (:obj:`False`)

        Returns
        -------
        out : np.ndarray
            The transformed features
        """
        pass


class Dither(PreProcessor):
    """Add random noise to a signal tensor

    The default axis of `apply` has been set to None, which will generate random noise
    for each coefficient. This is likely the desired behaviour. Setting axis to an
    integer will add random values along 1D slices of that axis.

    Intermediate values are calculated as 64-bit floats. The result is cast back to the
    input data type.

    Parameters
    ----------
    coeff
        Added noise will be in the range ``[-coeff, coeff]``
    """

    coeff: float  #:
    aliases = {"dither", "dithering"}  #:

    def __init__(self, coeff: float = 1.0):
        super().__init__()
        self.coeff = coeff

    def apply(
        self, signal: np.ndarray, axis: Optional[int] = None, in_place: bool = False
    ) -> np.ndarray:
        signal_dtype = signal.dtype
        if not in_place or signal.dtype != np.float64:
            signal = signal.astype(np.float64)
        if axis is None or not signal.shape or len(signal.shape) == 1:
            signal += self.coeff * np.random.random(signal.shape)
        else:
            random_shape = [1] * len(signal.shape)
            random_shape[axis] = signal.shape[axis]
            signal += self.coeff * np.random.random(random_shape)
        return signal.astype(signal_dtype, copy=False)


class Preemphasize(PreProcessor):
    """Attenuate the low frequencies of a signal by taking sample differences

    The following transformation is applied along the target axis

    ::

        new[i] = old[i] - coeff * old[i-1] for i > 1
        new[0] = old[0]

    This is essentially a convolution with a Haar wavelet for positive `coeff`. It
    emphasizes high frequencies.

    Intermediate values are calculated as 64-bit floats. The result is cast back to the
    input data type.

    Parameters
    ----------
    coeff
    """

    coeff: float  #:
    aliases = {"preemphasize", "preemphasis", "preemph"}  #:

    def __init__(self, coeff: float = 0.97):
        super().__init__()
        self.coeff = coeff

    def apply(
        self, signal: np.ndarray, axis: int = -1, in_place: bool = False
    ) -> np.ndarray:
        signal_dtype = signal.dtype
        if not in_place or signal_dtype != np.float64:
            signal = signal.astype(np.float64)
        if signal.shape and len(signal.shape) > 1:
            signal[1:] -= self.coeff * signal[:-1]
        else:
            tensor_slice_mind = [slice(None)] * len(signal.shape)
            tensor_slice_subhd = [slice(None)] * len(signal.shape)
            tensor_slice_mind[axis] = slice(1, None)
            tensor_slice_subhd[axis] = slice(-1, None)
            signal[tensor_slice_mind] -= self.coeff * signal[tensor_slice_subhd]
        return signal.astype(signal_dtype, copy=False)



# Copyright 2021 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Package constants used throughout pydrobert.speech"""

from typing import Set

__all__ = [
    "EFFECTIVE_SUPPORT_THRESHOLD",
    "LOG_FLOOR_VALUE",
    "SOUNDFILE_SUPPORTED_FILE_TYPES",
    "USE_FFTPACK",
]


USE_FFTPACK = False
"""bool : Whether to use :mod:`scipy.fftpack`

The scipy implementation of the FFT can be much faster than the numpy one. This is set
automatically to :obj:`True` if :mod:`scipy.fftpack` can be imported. It can be set to
:obj:`False` to use the numpy implementation.

:meta hide-value:
"""
try:
    from scipy import fftpack  # noqa: F401

    USE_FFTPACK = True
except ImportError:
    pass

EFFECTIVE_SUPPORT_THRESHOLD = 5e-4
"""float : Value considered roughly zero for support computations

No function is compactly supported in both the time and Fourier domains, but large
regions of either domain can be very close to zero. This value serves as a threshold for
zero. The higher it is, the more accurate computations will be, but the longer they will
take
"""

LOG_FLOOR_VALUE = 1e-5
"""float : Value used as floor when taking log in computations"""


# N.B. libsndfile's sphere support currently can't decode as many sphere encodings
# as _sphere.py can
_BASE_SOUNDFILE_SUPPORTED_TYPES = {"wav", "ogg", "flac", "aiff"}
_FULL_SOUNDFILE_SUPPORTED_TYPES = set()

SOUNDFILE_SUPPORTED_FILE_TYPES: Set[str] = set()
"""set : A list of the types of files SoundFile will be responsible for reading

If :mod:`soundfile` can be imported, it's the intersection of
:func:`soundfile.available_formats` with the set "wav", "ogg", "flac", and "aiff".

:meta hide-value:

See Also
--------
pydrobert.speech.util.read_signal
    Where this flag is used
"""

try:
    import soundfile as sf

    _FULL_SOUNDFILE_SUPPORTED_TYPES = set(x.lower() for x in sf.available_formats())

    SOUNDFILE_SUPPORTED_FILE_TYPES = (
        _BASE_SOUNDFILE_SUPPORTED_TYPES & _FULL_SOUNDFILE_SUPPORTED_TYPES
    )

except ImportError:
    pass




# Generate CLI page
# Needs imports, so run with full

import pydrobert.speech.command_line as cli
import os
from io import StringIO
import sys
import inspect
import warnings

warnings.simplefilter("ignore")

# Modified from
# https://stackoverflow.com/questions/16571150/how-to-capture-stdout-output-from-a-python-function-call
class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        sys.stdout = sys.stderr = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio  # free up some memory
        sys.stdout = self._stdout
        sys.stderr = self._stderr


DIR = os.path.dirname(__file__)
CLI_RST = os.path.join(DIR, "source", "cli.rst")

buff = "Command-Line Interface\n======================\n\n"
for cmd_name in ("compute-feats-from-kaldi-tables", "signals-to-torch-feat-dir"):
    buff += cmd_name + "\n" + ("-" * len(cmd_name)) + "\n\n::\n\n  "
    sys.argv[0] = cmd_name
    func = next(
        x[1] for x in inspect.getmembers(cli) if x[0] == cmd_name.replace("-", "_")
    )
    with Capturing() as c:
        try:
            func(["-h"])
        except SystemExit:
            pass
    buff += "\n  ".join(c) + "\n\n"

with open(CLI_RST, "w") as f:
    f.write(buff)



import os

from json import load as json_load

import numpy as np
import pydrobert.speech.compute as compute
import pydrobert.speech.config as config
import pytest

from pydrobert.speech.alias import alias_factory_subclass_from_arg
from pickle import load as pickle_load


@pytest.fixture(
    params=["causal", "centered"], scope="module",
)
def frame_style(request):
    return request.param


@pytest.fixture(
    params=[
        lambda frame_style: compute.STFTFrameComputer(
            {"name": "gabor", "scaling_function": "mel"},
            frame_length_ms=25,
            frame_shift_ms=10,
            use_power=True,
            use_log=True,
            pad_to_nearest_power_of_two=np.random.randint(2),
            include_energy=np.random.randint(2),
            frame_style=frame_style,
        ),
        lambda frame_style: compute.SIFrameComputer(
            {"name": "gabor", "scaling_function": "mel"},
            frame_shift_ms=25,
            use_power=True,
            use_log=True,
            pad_to_nearest_power_of_two=np.random.randint(2),
            include_energy=np.random.randint(2),
            frame_style=frame_style,
        ),
    ],
    ids=["stft", "si"],
)
def computer(request, frame_style):
    return request.param(frame_style)


@pytest.fixture(
    params=[0, 1, 2 ** 8, 2 ** 10],
    ids=["empty buffer", "length 1 buffer", "medium buffer", "large buffer"],
    scope="module",
)
def buff(request):
    b = np.random.random(request.param)
    b.flags.writeable = False
    return b


def test_framewise_matches_full(computer, buff):
    feats_full = computer.compute_full(buff)
    feats_framewise = compute.frame_by_frame_calculation(computer, buff)
    assert np.allclose(feats_full, feats_framewise), (
        feats_full.shape[0],
        np.where(np.logical_not(np.isclose(feats_full, feats_framewise)))[0],
    )


def test_chunk_sizes_dont_matter_to_result(computer, buff):
    feats = compute.frame_by_frame_calculation(computer, buff)
    feats_chunks = []
    while len(buff):
        next_len = np.random.randint(len(buff) + 1)
        feats_chunks.append(computer.compute_chunk(buff[:next_len]))
        buff = buff[next_len:]
    feats_chunks.append(computer.finalize())
    assert np.allclose(feats, np.concatenate(feats_chunks)), (
        feats.shape[0],
        np.where(np.logical_not(np.isclose(feats, np.concatenate(feats_chunks)))),
    )


def test_zero_samples_generate_zero_features(computer):
    assert computer.compute_full(np.empty(0)).shape == (0, computer.num_coeffs)
    assert computer.compute_chunk(np.empty(0)).shape == (0, computer.num_coeffs)
    assert computer.finalize().shape == (0, computer.num_coeffs)


def test_finalize_twice_generates_no_coefficients(computer):
    buff = np.random.random(computer.frame_length * 2)
    buff.flags.writeable = False
    coeffs = np.concatenate([computer.compute_chunk(buff), computer.finalize()])
    assert coeffs.shape[0] >= 1
    assert computer.finalize().shape == (0, computer.num_coeffs)


def test_started_makes_sense(computer):
    assert not computer.started
    computer.compute_chunk(np.empty(1))
    assert computer.started
    computer.finalize()
    assert not computer.started


def test_repeated_calls_generate_same_results(computer, buff):
    assert np.allclose(computer.compute_full(buff), computer.compute_full(buff))
    assert np.allclose(
        compute.frame_by_frame_calculation(computer, buff),
        compute.frame_by_frame_calculation(computer, buff),
    )


class TestFFTPACK(object):
    def setup_method(self):
        pytest.importorskip("scipy")
        self._orig_USE_FFTPACK = config.USE_FFTPACK

    def test_computations_same_between_numpy_scipy(self, computer, buff):
        config.USE_FFTPACK = False
        np_feats = computer.compute_full(buff)
        config.USE_FFTPACK = True
        scipy_feats = computer.compute_full(buff)
        assert np.allclose(np_feats, scipy_feats)

    def teardown_method(self):
        config.USE_FFTPACK = self._orig_USE_FFTPACK


class SIFrameComputerNpConvolve(compute.SIFrameComputer):
    """For compute_full, use np.convolve instead of overlap-save"""

    aliases = {}

    def __init__(self, *args, **kwargs):
        super(SIFrameComputerNpConvolve, self).__init__(*args, **kwargs)
        # convert filters back to impulse responses...
        for filt_idx in range(len(self._filts)):
            ifilt = self._compute_idft(self._filts[filt_idx])
            assert np.allclose(ifilt[self._max_support + 1 :], 0)
            self._filts[filt_idx] = ifilt[: self._max_support]
        self._window = self._window.flatten()

    def compute_full(self, signal):
        frame_length = self._frame_length
        frame_shift = self._frame_shift
        num_frames = len(signal) + frame_shift // 2
        num_frames //= frame_shift
        coeffs = np.empty((num_frames, self.num_coeffs), dtype=signal.dtype)
        if not num_frames:
            return coeffs
        signal = np.pad(
            signal,
            (
                max(0, frame_shift - self._translation),
                frame_length + self._translation,  # liberal
            ),
            "constant",
        )
        for coeff_idx, filt in enumerate(self._filts):
            y = np.convolve(signal, filt)
            if self._power:
                y[:] = y * y.conj()
            else:
                y[:] = np.abs(y)
            if self._frame_style == "centered":
                frame_start = max(0, self._translation - frame_shift)
            else:
                frame_start = self._translation
            for frame_idx in range(num_frames):
                frame_end = frame_start + 2 * self._frame_shift
                frame = y[frame_start:frame_end]
                coeffs[frame_idx, coeff_idx] = np.sum(frame.real * self._window)
                frame_start += self._frame_shift
        if self._log:
            coeffs[:] = np.log(np.maximum(coeffs, config.LOG_FLOOR_VALUE))
        return coeffs


def test_overlap_save_convolve_the_same(buff):
    os_computer = compute.SIFrameComputer({"name": "gabor", "scaling_function": "mel"})
    conv_computer = SIFrameComputerNpConvolve(
        {"name": "gabor", "scaling_function": "mel"}
    )
    os_coeffs = os_computer.compute_full(buff)
    conv_coeffs = conv_computer.compute_full(buff)
    assert os_coeffs.shape == conv_coeffs.shape
    assert np.allclose(os_coeffs, conv_coeffs)


def test_kaldi_comp_matches_fbank_comp():
    data_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
    with open(os.path.join(data_dir, "kaldi_feats.pkl"), "rb") as pkl_file:
        kaldi_feats = pickle_load(pkl_file)
    with open(os.path.join(data_dir, "noise.pkl"), "rb") as pkl_file:
        sig = pickle_load(pkl_file)
    with open(os.path.join(data_dir, "fbank.json")) as json_file:
        computer = alias_factory_subclass_from_arg(
            compute.FrameComputer, json_load(json_file)
        )
    fbank_feats = computer.compute_full(sig)
    # remove unit-normalization of hann window. We use the power spectrum, so
    # the denominator was squared, hence the 2x in the log domain
    fbank_feats += 2 * np.log(0.5 * (computer.frame_length - 1))
    # removes x2 for real-spectrum output. We double because of the
    # hermitian symmetry of real-valued signals; kaldi doesn't
    fbank_feats -= np.log(2)
    assert fbank_feats.shape == kaldi_feats.shape
    assert np.allclose(fbank_feats, kaldi_feats)
